# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""NIH CXR14 (relabeled) DataModule for computer-aided diagnosis.

This dataset was extracted from the clinical PACS database at the National
Institutes of Health Clinical Center (USA) and represents 60% of all their
radiographs. It contains labels for 14 common radiological signs in this order:
cardiomegaly, emphysema, effusion, hernia, infiltration, mass, nodule,
atelectasis, pneumothorax, pleural thickening, pneumonia, fibrosis, edema and
consolidation. Training and validation data come from the **relabeled** version
created in the :cite:p:`rajpurkar_deep_2018` study. Test data uses the available annotations with
:cite:p:`wang_chestx-ray8_2017`.

* Database references:

  * Original data: :cite:p:`wang_chestx-ray8_2017` (contains 112'120 chest X-ray images) and
    up to 14 associated radiological findings.
  * Labels and split references: We use train and validation splits published
    at :cite:p:`rajpurkar_deep_2018`, that are `available here` <nih-cxr14-relabeled_>`_.
    These are **different compared to the file lists provided with the original
    :cite:p:`wang_chestx-ray8_2017` study** (train/val set: 86'523 samples; test set: 25'595
    samples; +2 missing samples which are not listed, making up 112'120
    samples). The splits at :cite:p:`rajpurkar_deep_2018`, which we copied in this library,
    contain 104'987 samples which were relabeled making up a training and a
    validation set containing 98'637 and 6'350 samples respectively.  Note the
    relabeling work provided by :cite:p:`rajpurkar_deep_2018` does not provide **test** set
    annotations (only training and validation).  Our test set then consists of
    all CXR8_ samples that were not relabled, and for which we reused the
    `original CXR8 annotations <cxr8_>`_ (7'133 samples).

.. important:: **Raw data organization**

    The CXR8_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least the directory
    "images/" with all the images of the database.

    The labels from :cite:p:`rajpurkar_deep_2018` (`available here <nih-cxr14-relabeled_>`_)
    are already incorporated in this library and do **not** need to be
    re-downloaded.

    The flag ``idiap_folder_structure`` makes the loader search for files
    named, e.g. ``images/00030621_006.png``, as
    ``images/00030/00030621_006.png``.

* Raw data input (on disk):

  * PNG RGB 8-bit depth images
  * Original resolution: 1024 x 1024 pixels
  * Non-exclusive labels organized in a (compact) string list encoded
    as such:

    0. ``car``: cardiomegaly
    1. ``emp``: emphysema
    2. ``eff``: effusion
    3. ``her``: hernia
    4. ``inf``: infiltration
    5. ``mas``: mass
    6. ``nod``: nodule
    7. ``ate``: atelectasis
    8. ``pnt``: pneumothorax
    9. ``plt``: pleural thickening
    10. ``pne``: pneumonia
    11. ``fib``: fibrosis
    12. ``ede``: edema
    13. ``con``: consolidation

  * Patient age (integer)
  * Patient gender ("M" or "F")
  * Total samples available: 112'120

* Output image:

  * Transforms:

    * Load raw PNG with :py:mod:`PIL`, with auto-conversion to grayscale
    * Convert to torch tensor

  * Final specifications:

    * RGB, encoded as a 3-plane tensor, 32-bit floats, square
      (1024x1024 px)

      This decoder loads this description and converts it to a binary
      multi-label representation.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import numpy
import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir.cxr8"
"""Key to search for in the configuration file for the root directory of this
database."""

CONFIGURATION_KEY_IDIAP_FILESTRUCTURE = "cxr8.idiap_folder_structure"
"""Key to search for in the configuration file indicating if the loader should
use standard or idiap-based file organisation structure.

It causes the internal loader to search for files in a slightly
different folder structure, that was adapted to Idiap's requirements
(number of files per folder to be less than 10k).
"""

RADIOLOGICAL_FINDINGS = [
    "car",  # cardiomegaly
    "emp",  # emphysema
    "eff",  # effusion
    "her",  # hernia
    "inf",  # infiltration
    "mas",  # mass
    "nod",  # nodule
    "ate",  # atelectasis
    "pnt",  # pneumothorax
    "plt",  # pleural thickening
    "pne",  # pneumonia
    "fib",  # fibrosis
    "ede",  # edema
    "con",  # consolidation
]
"""List of radiological findings (abbreviations) supported on this database."""

_RADIOLOGICAL_FINDINGS_MAPPINGS = {
    k: numpy.eye(1, len(RADIOLOGICAL_FINDINGS), i, dtype=float)[0]
    for i, k in enumerate(RADIOLOGICAL_FINDINGS)
}
"""Auto-calculated radiological findings mapping between strings and position vectors."""


def binarize_findings(lst: list[str]) -> torch.Tensor:
    """Binarize the input list of radiological findings.

    The output list contains zeros and ones, respecting the findings order in
    :py:data:`RADIOLOGICAL_FINDINGS`.

    Parameters
    ----------
    lst
        A list of radiological findings that will be converted.

    Returns
    -------
        A list containing a binarized version of the input list.
    """
    return torch.tensor(
        numpy.array(
            [numpy.zeros(len(RADIOLOGICAL_FINDINGS), dtype=float)]
            + [_RADIOLOGICAL_FINDINGS_MAPPINGS[k] for k in lst]
        ).sum(0),
    ).float()


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the NIH CXR-14 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    idiap_file_organisation: bool
    """If should use the Idiap's filesystem organisation when looking up data.

    This variable will be ``True``, if the user has set the configuration
    parameter ``nih_cxr14.idiap_file_organisation`` in the global configuration
    file.  It will cause internal loader to search for files in a slightly
    different folder structure, that was adapted to Idiap's requirements
    (number of files per folder to be less than 10k).
    """

    def __init__(self):
        rc = load_rc()
        self.datadir = pathlib.Path(
            rc.get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir)),
        )
        self.idiap_file_organisation = rc.get(
            CONFIGURATION_KEY_IDIAP_FILESTRUCTURE,
            False,
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        file_path = pathlib.Path(sample[0])  # default
        if self.idiap_file_organisation:
            # for folder lookup efficiency, data is split into subfolders
            # each original file is on the subfolder `f[:5]/f`, where f
            # is the original file basename
            file_path = pathlib.Path(
                file_path.parent / file_path.name[:5] / file_path.name
            )

        # N.B.: some NIH CXR-14 images are encoded as color PNGs with an alpha
        # channel.  Most, are grayscale PNGs
        image = PIL.Image.open(self.datadir / file_path).convert(mode="L")
        image = to_dtype(to_image(image), torch.float32, scale=True)
        image = tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        The raw representation contains zero to many (unique) instances of
        radiological findings listed at :py:data:`RADIOLOGICAL_FINDINGS`.  This
        list is binarized (into 14 binary ositions) before it is returned.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing the
            sample target.

        Returns
        -------
            The labels corresponding to all radiological signs present in the
            specified sample, encapsulated as a 1D torch float tensor.
        """

        return binarize_findings(sample[1])


class DataModule(CachingDataModule):
    """NIH CXR14 (relabeled) DataModule for computer-aided diagnosis.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=14,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import json
from datetime import datetime

import click
from clapper.click import ConfigCommand, verbosity_option

from .logging import setup_cli_logger
from .train import reusable_options as training_options

# avoids X11/graphical desktop requirement when creating plots
__import__("matplotlib").use("agg")

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

\b
  1. Train a lwnet model with drive dataset, on the CPU, for only two
     epochs, then runs prediction (inference) and evaluation on listed datasets,
     report performance as a table and figures:

     .. code:: sh

        $ mednet experiment -vv lwnet drive --epochs=2
""",
)
@training_options
@verbosity_option(logger=logger, expose_value=False)
@click.pass_context
def experiment(
    ctx,
    model,
    output_folder,
    epochs,
    batch_size,
    accumulate_grad_batches,
    drop_incomplete_batch,
    datamodule,
    validation_period,
    device,
    cache_samples,
    seed,
    parallel,
    monitoring_interval,
    augmentations,
    balance_classes,
    initial_weights,
    **_,
):  # numpydoc ignore=PR01
    """Run a complete experiment, from training, to prediction and evaluation.

    Accepts all of command-line options from the ``train`` script.

    This script is just a wrapper around the individual scripts for training,
    running prediction, and evaluating.  It organises the output in a preset way:

    .. code::

       └─ <output-folder>/  # the generated model will be here
          ├── predictions.json  # the prediction outputs
          └── evaluation.json  # the evaluation outputs

    Note that complete experiments do not have options related to specific task
    functionality. Only the most generic set of options is available. To
    execute specific tasks, use the specialized command-line interfaces
    available in task-specific command groups.
    """

    experiment_start_timestamp = datetime.now()

    train_start_timestamp = datetime.now()
    logger.info(f"Started training at {train_start_timestamp}")

    from .train import train

    ctx.invoke(
        train,
        model=model,
        output_folder=output_folder,
        epochs=epochs,
        batch_size=batch_size,
        accumulate_grad_batches=accumulate_grad_batches,
        drop_incomplete_batch=drop_incomplete_batch,
        datamodule=datamodule,
        validation_period=validation_period,
        device=device,
        cache_samples=cache_samples,
        seed=seed,
        parallel=parallel,
        monitoring_interval=monitoring_interval,
        augmentations=augmentations,
        balance_classes=balance_classes,
        initial_weights=initial_weights,
    )
    train_stop_timestamp = datetime.now()

    logger.info(f"Ended training in {train_stop_timestamp}")
    logger.info(f"Training runtime: {train_stop_timestamp-train_start_timestamp}")

    logger.info("Started train analysis")
    from .train_analysis import train_analysis

    logdir = output_folder / "logs"
    ctx.invoke(
        train_analysis,
        logdir=logdir,
        output_folder=output_folder,
    )

    logger.info("Ended train analysis")

    predict_start_timestamp = datetime.now()
    logger.info(f"Started prediction at {predict_start_timestamp}")

    from .predict import predict

    ctx.invoke(
        predict,
        output_folder=output_folder,
        model=model,
        datamodule=datamodule,
        batch_size=batch_size,
        device=device,
        weight=output_folder,
        parallel=parallel,
    )

    predict_stop_timestamp = datetime.now()
    logger.info(f"Ended prediction in {predict_stop_timestamp}")
    logger.info(f"Prediction runtime: {predict_stop_timestamp-predict_start_timestamp}")

    evaluation_start_timestamp = datetime.now()
    logger.info(f"Started evaluation at {evaluation_start_timestamp}")

    predictions_file = output_folder / "predictions.json"

    with (predictions_file).open() as pf:
        splits = json.load(pf).keys()

        if "validation" in splits:
            evaluation_threshold = "validation"
        elif "train" in splits:
            evaluation_threshold = "train"
        else:
            evaluation_threshold = None

    match datamodule.task:
        case "classification":
            from .classify.evaluate import evaluate

            ctx.invoke(
                evaluate,
                predictions=predictions_file,
                output_folder=output_folder,
                threshold=evaluation_threshold,
                plot=True,
            )

        case "segmentation":
            from .segment.evaluate import evaluate

            ctx.invoke(
                evaluate,
                predictions=predictions_file,
                output_folder=output_folder,
                threshold=evaluation_threshold,
                plot=True,
            )

        case "detection":
            from .detect.evaluate import evaluate

            ctx.invoke(
                evaluate,
                predictions=predictions_file,
                output_folder=output_folder,
            )

        case _:
            raise click.BadParameter(
                f"Do not know how to handle evaluation on `{datamodule.task}` "
                f"task from `{type(datamodule).__module__}.{type(datamodule).__name__}`"
            )

    evaluation_stop_timestamp = datetime.now()
    logger.info(f"Ended prediction in {evaluation_stop_timestamp}")
    logger.info(
        f"Prediction runtime: {evaluation_stop_timestamp-evaluation_start_timestamp}"
    )

    experiment_stop_timestamp = datetime.now()
    logger.info(
        f"Total experiment runtime: {experiment_stop_timestamp-experiment_start_timestamp}"
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Support for switching execution devices (GPU vs CPU)."""

import logging
import os
import typing

import torch
import torch.backends

logger = logging.getLogger(__name__)


SupportedPytorchDevice: typing.TypeAlias = typing.Literal[
    "cpu",
    "cuda",
    "mps",
]
"""List of supported pytorch devices by this library."""


def _split_int_list(s: str) -> list[int]:
    """Split a list of integers encoded in a string (e.g. "1,2,3") into a
    Python list of integers (e.g. ``[1, 2, 3]``).

    Parameters
    ----------
    s
        A list of integers encoded in a string.

    Returns
    -------
    list[int]
        A Python list of integers.
    """

    return [int(k.strip()) for k in s.split(",")]


class DeviceManager:
    r"""Manage Lightning Accelerator and Pytorch Devices.

    It takes the user input, in the form of a string defined by
    ``[\S+][:\d[,\d]?]?`` (e.g.: ``cpu``, ``mps``, or ``cuda:3``), and can
    translate to the right incarnation of Pytorch devices or Lightning
    Accelerators to interface with the various frameworks.

    Instances of this class also manage the environment variable
    ``$CUDA_VISIBLE_DEVICES`` if necessary.

    Parameters
    ----------
    name
        The name of the device to use, in the form of a string defined by
        ``[\S+][:\d[,\d]?]?`` (e.g.: ``cpu``, ``mps``, or ``cuda:3``).  In
        the specific case of ``cuda``, one can also specify a device to use
        either by adding ``:N``, where N is the zero-indexed board number on
        the computer, or by setting the environment variable
        ``$CUDA_VISIBLE_DEVICES`` with the devices that are usable by the
        current process.
    """

    def __init__(self, name: SupportedPytorchDevice):
        parts = name.split(":", 1)

        # make device type of the right Python type
        if parts[0] not in typing.get_args(SupportedPytorchDevice):
            raise ValueError(f"Unsupported device-type `{parts[0]}`")
        self.device_type: SupportedPytorchDevice = typing.cast(
            SupportedPytorchDevice,
            parts[0],
        )

        self.device_ids: list[int] = []
        if len(parts) > 1:
            self.device_ids = _split_int_list(parts[1])

        if self.device_type == "cuda":
            visible_env = os.environ.get("CUDA_VISIBLE_DEVICES")
            if visible_env:
                visible = _split_int_list(visible_env)
                if self.device_ids and visible != self.device_ids:
                    logger.warning(
                        f"${{CUDA_VISIBLE_DEVICES}}={visible} and name={name} "
                        f"- overriding environment with value set on `name`",
                    )
                else:
                    self.device_ids = visible

            # make sure that it is consistent with the environment
            if self.device_ids:
                os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(
                    [str(k) for k in self.device_ids],
                )

        if self.device_type not in typing.get_args(SupportedPytorchDevice):
            raise RuntimeError(
                f"Unsupported device type `{self.device_type}`. "
                f"Supported devices types are "
                f"`{', '.join(typing.get_args(SupportedPytorchDevice))}`",
            )

        if self.device_ids and self.device_type in ("cpu", "mps"):
            logger.warning(
                f"Cannot pin device ids if using cpu or mps backend. "
                f"Setting `name` to {name} is non-sensical.  Ignoring...",
            )

        # check if the device_type that was set has support compiled in
        if self.device_type == "cuda":
            assert hasattr(torch, "cuda") and torch.cuda.is_available(), (
                f"User asked for device = `{name}`, but CUDA support is "
                f"not compiled into pytorch!"
            )

        if self.device_type == "mps":
            assert (
                hasattr(torch.backends, "mps") and torch.backends.mps.is_available()  # type:ignore
            ), (
                f"User asked for device = `{name}`, but MPS support is "
                f"not compiled into pytorch!"
            )

    def torch_device(self) -> torch.device:
        """Return a representation of the torch device to use by default.

        .. warning::

           If a list of devices is set, then this method only returns the first
           device.  This may impact Nvidia GPU logging in the case multiple
           GPU cards are used.

        Returns
        -------
        torch.device
            The **first** torch device (if a list of ids is set).
        """

        if self.device_type in ("cpu", "mps"):
            return torch.device(self.device_type)

        if self.device_type == "cuda":
            if not self.device_ids:
                return torch.device(self.device_type)

            return torch.device(self.device_type, self.device_ids[0])

        # if you get to this point, this is an unexpected RuntimeError
        raise RuntimeError(
            f"Unexpected device type {self.device_type} lacks support",
        )

    def lightning_accelerator(self) -> tuple[str, int | list[int] | str]:
        """Return the lightning accelerator setup.

        Returns
        -------
        accelerator
            The lightning accelerator to use.
        devices
            The lightning devices to use.
        """

        devices: int | list[int] | str = self.device_ids
        if not devices:
            devices = "auto"
        elif self.device_type == "mps":
            devices = 1

        return self.device_type, devices



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`Indian database <mednet.data.classify.indian>` (cross validation fold 4)."""

import importlib.resources

from mednet.data.classify.indian import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0]) / "fold-4.json"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import pathlib
import re
import typing

logger = logging.getLogger(__name__)


CHECKPOINT_ALIASES = {
    "best": "model-at-lowest-validation-loss-{epoch}",
    "periodic": "model-at-{epoch}",
}
"""Standard paths where checkpoints may be (if produced with this
framework)."""

CHECKPOINT_EXTENSION = ".ckpt"


def _get_checkpoint_from_alias(
    path: pathlib.Path,
    alias: typing.Literal["best", "periodic"],
) -> pathlib.Path:
    """Get an existing checkpoint file path.

    This function can search for names matching the checkpoint alias "stem"
    (ie. the prefix), and then assumes a dash "-" and a number follows that
    prefix before the expected file extension.  The number is parsed and
    considred to be an epoch number.  The latest file (the file containing the
    highest epoch number) is returned.

    If only one file is present matching the alias characteristics, then it is
    returned.

    Parameters
    ----------
    path
        Folder in which may contain checkpoint.
    alias
        Can be one of "best" or "periodic".

    Returns
    -------
        Path to the requested checkpoint, or ``None``, if no checkpoint file
        matching specifications is found on the provided path.

    Raises
    ------
    FileNotFoundError
        In case it cannot find any file on the provided path matching the given
        specifications.
    """

    template = path / (CHECKPOINT_ALIASES[alias] + CHECKPOINT_EXTENSION)

    if template.exists():
        return template

    # otherwise, we see if we are looking for a template instead, in which case
    # we must pick the latest.
    assert "{epoch}" in str(
        template,
    ), f"Template `{str(template)}` does not contain the keyword `{{epoch}}`"

    pattern = re.compile(
        template.name.replace("{epoch}", r"epoch(?P<separator>=|-|_)(?P<epoch>\d+)"),
    )
    highest = -1
    separator = "="
    for f in template.parent.iterdir():
        match = pattern.match(f.name)
        if match is not None:
            value = int(match.group("epoch"))
            if value > highest:
                highest = value
                separator = match.group("separator")

    if highest != -1:
        return template.with_name(
            template.name.replace("{epoch}", f"epoch{separator}{highest}"),
        )

    raise FileNotFoundError(
        f"A file matching `{str(template)}` specifications was not found",
    )


def get_checkpoint_to_resume_training(
    path: pathlib.Path,
) -> pathlib.Path:
    """Return the best checkpoint file path to resume training from.

    Parameters
    ----------
    path
        The base directory containing either the "periodic" checkpoint to start
        the training session from.

    Returns
    -------
    pathlib.Path
        Path to a checkpoint file that exists on disk.

    Raises
    ------
    FileNotFoundError
        If none of the checkpoints can be found on the provided directory.
    """

    return _get_checkpoint_from_alias(path, "periodic")


def get_checkpoint_to_run_inference(
    path: pathlib.Path,
) -> pathlib.Path:
    """Return the best checkpoint file path to run inference with.

    Parameters
    ----------
    path
        The base directory containing either the "best", "last" or "periodic"
        checkpoint to start the training session from.

    Returns
    -------
    pathlib.Path
        Path to a checkpoint file that exists on disk.

    Raises
    ------
    FileNotFoundError
        If none of the checkpoints can be found on the provided directory.
    """

    try:
        return _get_checkpoint_from_alias(path, "best")
    except FileNotFoundError:
        logger.error(
            "Did not find lowest-validation-loss model to run inference "
            "from.  Trying to search for the last periodically saved model...",
        )

    return _get_checkpoint_from_alias(path, "periodic")



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Utilities for command-line scripts."""

import json
import logging
import pathlib
import re
import shutil
import typing

import compact_json
import lightning.pytorch
import lightning.pytorch.callbacks
import numpy
import torch.nn

from ..engine.device import SupportedPytorchDevice

logger = logging.getLogger(__name__)

JSONable: typing.TypeAlias = (
    typing.Mapping[str, "JSONable"]
    | typing.Sequence["JSONable"]
    | str
    | int
    | float
    | bool
    | None
)
"""Defines types that can be encoded in a JSON string."""


def model_summary(
    model: torch.nn.Module,
) -> dict[str, int | list[tuple[str, str, int]]]:
    """Save a little summary of the model in a txt file.

    Parameters
    ----------
    model
        Instance of the model for which to save the summary.

    Returns
    -------
    tuple[lightning.pytorch.callbacks.ModelSummary, int]
        A tuple with the model summary in a text format and number of parameters of the model.
    """

    s = lightning.pytorch.utilities.model_summary.ModelSummary(  # type: ignore
        model,
    )

    return dict(
        model_summary=list(zip(s.layer_names, s.layer_types, s.param_nums)),
        model_size=s.total_parameters,
    )


def device_properties(
    device_type: SupportedPytorchDevice,
) -> dict[str, int | float | str]:
    """Generate information concerning hardware properties.

    Parameters
    ----------
    device_type
        The type of compute device we are using.

    Returns
    -------
        Static properties of the current machine.
    """

    from ..utils.resources import cpu_constants, cuda_constants, mps_constants

    retval: dict[str, int | float | str] = {}
    retval.update(cpu_constants())

    match device_type:
        case "cpu":
            pass
        case "cuda":
            results = cuda_constants()
            if results is not None:
                retval.update(results)
        case "mps":
            results = mps_constants()
            if results is not None:
                retval.update(results)
        case _:
            pass

    return retval


def execution_metadata() -> dict[str, int | float | str | dict[str, str] | list[str]]:
    """Produce metadata concerning the running script, in the form of a
    dictionary.

    This function returns potentially useful metadata concerning program
    execution.  It contains a certain number of preset variables.

    Returns
    -------
        A dictionary that contains the following fields:

        * ``package-name``: current package name (e.g. ``mednet``)
        * ``package-version``: current package version (e.g. ``1.0.0b0``)
        * ``datetime``: date and time in ISO8601 format (e.g. ``2024-02-23T18:38:09+01:00``)
        * ``user``: username (e.g. ``johndoe``)
        * ``conda-env``: if set, the name of the current conda environment
        * ``path``: current path when executing the command
        * ``command-line``: the command-line that is being run
        * ``hostname``: machine hostname (e.g. ``localhost``)
        * ``platform``: machine platform (e.g. ``darwin``)
        * ``accelerator``: acceleration devices available (e.g. ``cuda``)
    """

    import datetime
    import importlib.metadata
    import importlib.util
    import os
    import sys

    args: list[str] = []
    for k in sys.argv:
        if " " in k:
            args.append(f"'{k}'")
        else:
            args.append(k)

    # current date time, in ISO8610 format
    current_datetime = datetime.datetime.now().astimezone().isoformat()

    # collects dependency information
    package_name = __package__.split(".")[0] if __package__ is not None else "unknown"
    requires = importlib.metadata.requires(package_name) or []
    dependence_names = [re.split(r"(\=|~|!|>|<|;|\s)+", k)[0] for k in requires]
    installed = {
        v[0]: k for k, v in importlib.metadata.packages_distributions().items()
    }
    dependencies = {
        k: importlib.metadata.version(k)  # version number as str
        for k in sorted(dependence_names)
        if importlib.util.find_spec(k if k not in installed else installed[k])
        is not None  # if is installed
    }

    # checks if the current version corresponds to a dirty (uncommitted) change
    # set, issues a warning to the user
    current_version = importlib.metadata.version(package_name)
    try:
        import versioningit

        actual_version = versioningit.get_version(".")
        if current_version != actual_version:
            logger.warning(
                f"Version mismatch between current version set "
                f"({current_version}) and actual version returned by "
                f"versioningit ({actual_version}).  This typically happens "
                f"when you commit changes locally and do not re-install the "
                f"package. Run `pixi update {package_name}`, `pip install -e .` "
                f"or equivalent to fix this.",
            )
    except Exception as e:
        # not in a git repo?
        logger.debug(f"Error {e}")
        pass

    # checks if any acceleration device is present in the current platform
    accelerators = [f"cpu ({torch.backends.cpu.get_cpu_capability()})"]

    if torch.cuda.is_available() and torch.backends.cuda.is_built():
        accelerators.append("cuda")

    if torch.backends.cudnn.is_available():
        accelerators.append("cudnn")

    if torch.backends.mps.is_available():
        accelerators.append("mps")

    if torch.backends.mkl.is_available():
        accelerators.append("mkl")

    if torch.backends.mkldnn.is_available():
        accelerators.append("mkldnn")

    if torch.backends.openmp.is_available():
        accelerators.append("openmp")

    python = {
        "version": ".".join([str(k) for k in sys.version_info[:3]]),
        "path": sys.executable,
    }

    return {
        "datetime": current_datetime,
        "package-name": package_name,
        "package-version": current_version,
        "python": python,
        "dependencies": dependencies,
        "user": __import__("getpass").getuser(),
        "conda-env": os.environ.get("CONDA_DEFAULT_ENV", ""),
        "path": os.path.realpath(os.curdir),
        "command-line": " ".join(args),
        "hostname": __import__("platform").node(),
        "platform": sys.platform,
        "accelerators": accelerators,
    }


class NumpyJSONEncoder(json.JSONEncoder):
    """Extends the standard JSON encoder to support Numpy arrays."""

    def default(self, o: typing.Any) -> typing.Any:
        """If input object is a ndarray it will be converted into a list.

        Parameters
        ----------
        o
            Input object to be JSON serialized.

        Returns
        -------
            A serializable representation of object ``o``.
        """

        if isinstance(o, numpy.ndarray):
            try:
                retval = o.tolist()
            except TypeError:
                pass
            else:
                return retval
        elif isinstance(o, numpy.generic):
            try:
                retval = o.item()
            except TypeError:
                pass
            else:
                return retval

        # Let the base class default method raise the TypeError
        return super().default(o)


def save_json_with_backup(path: pathlib.Path, data: JSONable) -> None:
    """Save a dictionary into a JSON file with path checking and backup.

    This function will save a dictionary into a JSON file.  It will check to
    the existence of the directory leading to the file and create it if
    necessary.  If the file already exists on the destination folder, it is
    backed-up before a new file is created with the new contents.

    Parameters
    ----------
    path
        The full path where to save the JSON data.
    data
        The data to save on the JSON file.
    """

    formatter = compact_json.Formatter()
    # only only 2 indent spaces for further levels
    formatter.indent_spaces = 2
    # controls how much nesting can happen
    formatter.max_inline_complexity = 2
    # controls the maximum line width (has priority over nesting)
    formatter.max_inline_length = 88
    # remove any trailing whitespaces
    formatter.omit_trailing_whitespace = True

    path.parent.mkdir(parents=True, exist_ok=True)
    if path.exists():
        backup = path.parent / (path.name + "~")
        shutil.copy(path, backup)

    data = json.loads(json.dumps(data, indent=2, cls=NumpyJSONEncoder))
    formatter.dump(data, str(path))


def save_json_metadata(
    output_file: pathlib.Path,
    **kwargs: typing.Any,
) -> None:  # numpydoc ignore=PR01
    """Save prediction hyperparameters into a .json file."""

    from ..data.datamodule import ConcatDataModule
    from ..engine.device import DeviceManager
    from ..models.model import Model
    from .utils import (
        device_properties,
        execution_metadata,
        model_summary,
        save_json_with_backup,
    )

    json_data: dict[str, typing.Any] = execution_metadata()

    for key, value in kwargs.items():
        match value:
            case ConcatDataModule():
                json_data["database_name"] = value.database_name
                json_data["database_split"] = value.split_name
            case Model():
                json_data["model"] = f"{type(value).__module__}.{type(value).__name__}"
                json_data.update(model_summary(value))
            case pathlib.Path():
                json_data[key] = str(value)
            case DeviceManager():
                json_data.update(device_properties(value.device_type))
            case list() if key == "augmentations":
                if len(value) != 0:
                    json_data[key] = [f"{type(k).__module__}.{str(k)}" for k in value]
                else:
                    json_data[key] = []
            case _:
                json_data[key] = value

    json_data = {k.replace("_", "-"): v for k, v in json_data.items()}
    logger.info(f"Writing run metadata at `{output_file}`...")
    save_json_with_backup(output_file, json_data)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""ChestX-ray8: Hospital-scale Chest X-ray Database.

The database contains a total of 112'120 images. Image size for each X-ray is
1024 x 1024. One set of automatically generated mask annotations is available
for all images.

* Database references:

  * Original data: :cite:p:`wang_chestx-ray8_2017`
  * Split reference: :cite:p:`gaal_attention_2020`

.. important:: **Raw data organization**

    The CXR8_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least the following
    directories:

    - ``images/`` (directory containing the CXR images, in PNG format)
    - ``segmentations/`` (must contain masks downloaded from `CXR8-Annotations`_)

    The flag ``idiap_folder_structure`` makes the loader search for files
    named, e.g. ``images/00030621_006.png``, as
    ``images/00030/00030621_006.png`` (this is valid for both images and
    segmentation masks).

* Raw data input (on disk):

  * PNG RGB 8-bit depth images
  * Resolution: 1024 x 1024 pixels
  * Total samples available: 112'120

* Output image:

  * Transforms:

    * Load raw PNG with :py:mod:`PIL`, with auto-conversion to RGB, convert to
      tensor
    * Labels for each of the lungs are read from the provided GIF files and
      merged into a single output image.

The ``default`` split contains 78'484 images for training, 11'212 images for
validation, and 22'424 images for testing.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""

CONFIGURATION_KEY_IDIAP_FILESTRUCTURE = DATABASE_SLUG + ".idiap_folder_structure"
"""Key to search for in the configuration file indicating if the loader should
use standard or idiap-based file organisation structure.

It causes the internal loader to search for files in a slightly different
folder structure, that was adapted to Idiap's requirements (number of files per
folder to be less than 10k).
"""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the cxr8 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )
        self.idiap_file_organisation = load_rc().get(
            CONFIGURATION_KEY_IDIAP_FILESTRUCTURE,
            False,
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        file_path = pathlib.Path(sample[0])

        if self.idiap_file_organisation:
            sample_parts = sample[0].split("/", 1)
            file_path = pathlib.Path(
                sample_parts[0] + "/" + sample_parts[1][:5] + "/" + sample_parts[1]
            )

        image = PIL.Image.open(self.datadir / file_path).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        # use image as a base since target() can be overriden by child class
        mask = torch.ones((1, image.shape[-2], image.shape[-1]), dtype=torch.float32)

        image = tv_tensors.Image(image)
        target = tv_tensors.Mask(target)
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """
        target_path = pathlib.Path(sample[1])

        if self.idiap_file_organisation:
            target_parts = sample[1].split("/", 1)
            target_path = pathlib.Path(
                pathlib.Path(
                    target_parts[0] + "/" + target_parts[1][:5] + "/" + target_parts[1]
                )
            )

        target = PIL.Image.open(self.datadir / target_path).convert(
            mode="1", dither=None
        )
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """ChestX-ray8: Hospital-scale Chest X-ray Database.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines most common types used in code."""

import typing

Checkpoint: typing.TypeAlias = typing.MutableMapping[str, typing.Any]
"""Definition of a lightning checkpoint."""

TaskType: typing.TypeAlias = typing.Literal["classification", "segmentation"]
"""Types of supported tasks."""

TargetType: typing.TypeAlias = typing.Literal["binary", "multiclass", "multilabel"]
"""Types of classifiers/segmenters we support.

We distinguish target types by looking at sample target values of available databases.

* ``binary``: single (binary) target, where negative labels are set to zero and
   positives to 1.
* ``multiclass``: multiple (binary) outputs, one-hot encoded where
  negative/positive are encoded as in the binary case. Each sample is assigned
  exactly one, and only one, label.
* ``multilabel``: multiple outputs, where each sample may have more than one
  label.
"""



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Definition of types related to models used in object detection tasks."""

import typing

Prediction: typing.TypeAlias = tuple[
    str,
    typing.Sequence[tuple[typing.Sequence[int], int]],
    typing.Sequence[tuple[typing.Sequence[int | float], int, float]],
]
"""The sample name, the target (boxes and labels), and the predicted value (boxes,
labels and scores)."""

PredictionSplit: typing.TypeAlias = typing.Mapping[str, typing.Sequence[Prediction]]
"""A series of predictions for different database splits."""



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Defines functionality for the evaluation of classification predictions."""

import logging
import typing
from collections.abc import Iterable

import credible.bayesian.metrics
import credible.curves
import credible.plot
import matplotlib.axes
import matplotlib.figure
import numpy
import numpy.typing
import sklearn.metrics
import tabulate
from matplotlib import pyplot as plt

from ...models.classify.typing import Prediction

logger = logging.getLogger(__name__)


def eer_threshold(predictions: Iterable[Prediction]) -> float:
    """Calculate the (approximate) threshold leading to the equal error rate.

    For multi-label problems, calculate the EER threshold in the "micro" sense
    by first rasterizing all scores and labels (with :py:func:`numpy.ravel`),
    and then using this (large) 1D vector like in a binary classifier.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.classify.typing.Prediction`'s.

    Returns
    -------
    float
        The EER threshold value.
    """

    from scipy.interpolate import interp1d
    from scipy.optimize import brentq

    y_scores = numpy.array([k[2] for k in predictions]).ravel()
    y_labels = numpy.array([k[1] for k in predictions]).ravel()

    fpr, tpr, thresholds = sklearn.metrics.roc_curve(y_labels, y_scores)

    eer = brentq(lambda x: 1.0 - x - interp1d(fpr, tpr)(x), 0.0, 1.0)
    return float(interp1d(fpr, thresholds)(eer))


def _get_centered_maxf1(
    f1_scores: numpy.typing.NDArray,
    thresholds: numpy.typing.NDArray,
) -> tuple[float, float]:
    """Return the centered max F1 score threshold when multiple thresholds give
    the same max F1 score.

    Parameters
    ----------
    f1_scores
        1D array of f1 scores.
    thresholds
        1D array of thresholds.

    Returns
    -------
    tuple(float, float)
        A tuple with the maximum F1-score and the "centered" threshold.
    """

    maxf1 = f1_scores.max()
    maxf1_indices = numpy.where(f1_scores == maxf1)[0]

    # If multiple thresholds give the same max F1 score
    if len(maxf1_indices) > 1:
        mean_maxf1_index = int(round(numpy.mean(maxf1_indices)))
    else:
        mean_maxf1_index = maxf1_indices[0]

    return maxf1, thresholds[mean_maxf1_index]


def maxf1_threshold(predictions: Iterable[Prediction]) -> float:
    """Calculate the threshold leading to the maximum F1-score on a precision-
    recall curve.

    For multi-label problems, calculate the maximum F1-core threshold in the
    "micro" sense by first rasterizing all scores and labels (with
    :py:func:`numpy.ravel`), and then using this (large) 1D vector like in a
    binary classifier.

    Parameters
    ----------
    predictions
        An iterable of multiple
        :py:data:`.models.classify.typing.Prediction`'s.

    Returns
    -------
    float
        The threshold value leading to the maximum F1-score on the provided set
        of predictions.
    """

    y_scores = numpy.array([k[2] for k in predictions]).ravel()
    y_labels = numpy.array([k[1] for k in predictions]).ravel()

    precision, recall, thresholds = sklearn.metrics.precision_recall_curve(
        y_labels,
        y_scores,
    )

    numerator = 2 * recall * precision
    denom = recall + precision
    f1_scores = numpy.divide(
        numerator,
        denom,
        out=numpy.zeros_like(denom),
        where=(denom != 0),
    )

    _, maxf1_threshold = _get_centered_maxf1(f1_scores, thresholds)
    return maxf1_threshold


def run(
    name: str,
    predictions: typing.Sequence[Prediction],
    binning: str | int,
    threshold_a_priori: float | None = None,
    credible_regions: bool = False,
) -> dict[str, typing.Any]:
    """Run inference and calculates measures for binary or multilabel classification.

    For multi-label problems, calculate the metrics in the "micro" sense by
    first rasterizing all scores and labels (with :py:func:`numpy.ravel`), and
    then using this (large) 1D vector like in a binary classifier.

    Parameters
    ----------
    name
        The name of subset to load.
    predictions
        A list of predictions to consider for measurement.
    binning
        The binning algorithm to use for computing the bin widths and
        distribution for histograms.  Choose from algorithms supported by
        :py:func:`numpy.histogram`.
    threshold_a_priori
        A threshold to use, evaluated *a priori*, if must report single values.
        If this value is not provided, an *a posteriori* threshold is calculated
        on the input scores.  This is a biased estimator.
    credible_regions
        If set to ``True``, then returns also credible intervals via
        :py:mod:`credible.bayesian.metrics`. Notice the evaluation of ROC-AUC
        and Average Precision confidence margins can be rather slow for larger
        datasets.

    Returns
    -------
    dict[str, typing.Any]
        A dictionary containing the performance summary on the specified threshold,
        general performance curves (under the key ``curves``), and score histograms
        (under the key ``score-histograms``).
    """

    y_scores = numpy.array([k[2] for k in predictions]).ravel()
    y_labels = numpy.array([k[1] for k in predictions])
    ## ctype = classifier_type(y_labels)
    num_samples, num_classes = y_labels.shape
    y_labels = y_labels.ravel()

    neg_label = y_labels.min()
    pos_label = y_labels.max()

    use_threshold = threshold_a_priori
    if use_threshold is None:
        use_threshold = maxf1_threshold(predictions)
        logger.warning(
            f"User did not pass an *a priori* threshold for the evaluation "
            f"of split `{name}`.  Using threshold a posteriori (biased) with value "
            f"`{use_threshold:.4f}`",
        )

    y_predictions = numpy.where(y_scores >= use_threshold, pos_label, neg_label)

    summary = dict(
        num_samples=num_samples,
        num_classes=num_classes,
        threshold=use_threshold,
        threshold_a_posteriori=(threshold_a_priori is None),
        precision=sklearn.metrics.precision_score(
            y_labels, y_predictions, pos_label=pos_label
        ),
        recall=sklearn.metrics.recall_score(
            y_labels, y_predictions, pos_label=pos_label
        ),
        f1=sklearn.metrics.f1_score(y_labels, y_predictions, pos_label=pos_label),
        average_precision=sklearn.metrics.average_precision_score(
            y_labels, y_scores, pos_label=pos_label
        ),
        specificity=sklearn.metrics.recall_score(
            y_labels, y_predictions, pos_label=neg_label
        ),
        roc_auc=sklearn.metrics.roc_auc_score(y_labels, y_scores),
        accuracy=sklearn.metrics.accuracy_score(y_labels, y_predictions),
    )

    if credible_regions:
        logger.info(
            f"Computing credible regions for metrics on split `{name}` "
            f"(samples = {len(predictions)}) - "
            f"note this can be slow on very large datasets..."
        )
        f1 = credible.bayesian.metrics.f1_score(y_labels, y_predictions)
        roc_auc = credible.bayesian.metrics.roc_auc_score(y_labels, y_scores)
        precision = credible.bayesian.metrics.precision_score(y_labels, y_predictions)
        recall = credible.bayesian.metrics.recall_score(y_labels, y_predictions)
        average_precision = credible.bayesian.metrics.average_precision_score(
            y_labels, y_scores
        )
        specificity = credible.bayesian.metrics.specificity_score(
            y_labels, y_predictions
        )
        accuracy = credible.bayesian.metrics.accuracy_score(y_labels, y_predictions)

        summary.update(
            dict(
                precision_mean=precision[0],
                precision_mode=precision[1],
                precision_lo=precision[2],
                precision_hi=precision[3],
                recall_mean=recall[0],
                recall_mode=recall[1],
                recall_lo=recall[2],
                recall_hi=recall[3],
                f1_mean=f1[0],
                f1_mode=f1[1],
                f1_lo=f1[2],
                f1_hi=f1[3],
                average_precision_exact=average_precision[0],
                average_precision_lo=average_precision[1],
                average_precision_hi=average_precision[2],
                specificity_mean=specificity[0],
                specificity_mode=specificity[1],
                specificity_lo=specificity[2],
                specificity_hi=specificity[3],
                roc_auc_exact=roc_auc[0],
                roc_auc_lo=roc_auc[1],
                roc_auc_hi=roc_auc[2],
                accuracy_mean=accuracy[0],
                accuracy_mode=accuracy[1],
                accuracy_lo=accuracy[2],
                accuracy_hi=accuracy[3],
            )
        )

    # curves: ROC and precision recall
    summary["curves"] = dict(
        roc=dict(
            zip(
                ("fpr", "tpr", "thresholds"),
                sklearn.metrics.roc_curve(
                    y_labels,
                    y_scores,
                    pos_label=pos_label,
                ),
            ),
        ),
        precision_recall=dict(
            zip(
                ("precision", "recall", "thresholds"),
                sklearn.metrics.precision_recall_curve(
                    y_labels,
                    y_scores,
                    pos_label=pos_label,
                ),
            ),
        ),
    )

    # score histograms
    # what works: <integer>, doane*, scott, stone, rice*, sturges*, sqrt
    # what does not work: auto, fd
    summary["score-histograms"] = dict(
        positives=dict(
            zip(
                ("hist", "bin_edges"),
                numpy.histogram(
                    y_scores[y_labels == pos_label],
                    bins=binning,
                    range=(0, 1),
                ),
            ),
        ),
        negatives=dict(
            zip(
                ("hist", "bin_edges"),
                numpy.histogram(
                    y_scores[y_labels == neg_label],
                    bins=binning,
                    range=(0, 1),
                ),
            ),
        ),
    )

    return summary


def make_table(
    data: typing.Mapping[str, typing.Mapping[str, typing.Any]],
    fmt: str,
) -> str:
    """Tabulate summaries from multiple splits.

    This function can properly tabulate the various summaries produced for all
    the splits in a prediction database.

    Parameters
    ----------
    data
        An iterable over all summary data collected.
    fmt
        One of the formats supported by `python-tabulate
        <https://pypi.org/project/tabulate/>`_.

    Returns
    -------
    str
        A string containing the tabulated information.
    """

    def _exclusion_condition(v: str) -> bool:
        return not (
            v in ("curves", "score-histograms")
            or v.endswith(("_mean", "_mode", "_hi", "_lo", "_exact"))
        )

    # dump evaluation results in RST format to screen and file
    table_data = {}
    for k, v in data.items():
        table_data[k] = {kk: vv for kk, vv in v.items() if _exclusion_condition(kk)}

    example = next(iter(table_data.values()))
    headers = list(example.keys())
    table = [[k[h] for h in headers] for k in table_data.values()]

    # add subset names
    headers = ["subset"] + headers
    table = [[name] + k for name, k in zip(table_data.keys(), table)]

    return tabulate.tabulate(table, headers, tablefmt=fmt, floatfmt=".3f")


def _score_plot(
    histograms: dict[str, dict[str, numpy.typing.NDArray]],
    title: str,
    threshold: float | None,
) -> matplotlib.figure.Figure:
    """Plot the normalized score distributions for all systems.

    Parameters
    ----------
    histograms
        A dictionary containing all histograms that should be inserted into the
        plot.  Each histogram should itself be setup as another dictionary
        containing the keys ``hist`` and ``bin_edges`` as returned by
        :py:func:`numpy.histogram`.
    title
        Title of the plot.
    threshold
        Shows where the threshold is in the figure.  If set to ``None``, then
        does not show the threshold line.

    Returns
    -------
    matplotlib.figure.Figure
        A single (matplotlib) plot containing the score distribution, ready to
        be saved to disk or displayed.
    """

    from matplotlib.ticker import MaxNLocator

    fig, ax = plt.subplots(1, 1)
    assert isinstance(fig, matplotlib.figure.Figure)
    ax = typing.cast(matplotlib.axes.Axes, ax)  # gets editor to behave

    # Here, we configure the "style" of our plot
    ax.set_xlim((0, 1))
    ax.set_title(title)
    ax.set_xlabel("Score")
    ax.set_ylabel("Count")

    # Only show ticks on the left and bottom spines
    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)
    ax.get_xaxis().tick_bottom()
    ax.get_yaxis().tick_left()
    ax.get_yaxis().set_major_locator(MaxNLocator(integer=True))

    # Setup the grid
    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.2)
    ax.get_xaxis().grid(False)

    max_hist = 0
    for name in histograms.keys():
        hist = histograms[name]["hist"]
        bin_edges = histograms[name]["bin_edges"]
        width = 0.7 * (bin_edges[1] - bin_edges[0])
        center = (bin_edges[:-1] + bin_edges[1:]) / 2
        ax.bar(center, hist, align="center", width=width, label=name, alpha=0.7)
        max_hist = max(max_hist, hist.max())

    # Detach axes from the plot
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015 * max_hist))

    if threshold is not None:
        # Adds threshold line (dotted red)
        ax.axvline(
            threshold,  # type: ignore
            color="red",
            lw=2,
            alpha=0.75,
            ls="dotted",
            label="threshold",
        )

    # Adds a nice legend
    ax.legend(
        fancybox=True,
        framealpha=0.7,
    )

    # Makes sure the figure occupies most of the possible space
    fig.tight_layout()

    return fig


def make_plots(results: dict[str, dict[str, typing.Any]]) -> list:
    """Create plots for all curves and score distributions in ``results``.

    Parameters
    ----------
    results
        Evaluation data as returned by :py:func:`run`.

    Returns
    -------
        A list of figures to record to file
    """

    retval = []

    with credible.plot.tight_layout(
        ("False Positive Rate", "True Positive Rate"), "ROC"
    ) as (fig, ax):
        for split_name, data in results.items():
            _auroc = credible.curves.area_under_the_curve(
                (data["curves"]["roc"]["fpr"], data["curves"]["roc"]["tpr"]),
            )
            ax.plot(
                data["curves"]["roc"]["fpr"],
                data["curves"]["roc"]["tpr"],
                label=f"{split_name} (AUC: {_auroc:.2f})",
            )
        ax.legend(loc="best", fancybox=True, framealpha=0.7)
        retval.append(fig)

    with credible.plot.tight_layout_f1iso(
        ("Recall", "Precision"), "Precison-Recall"
    ) as (fig, ax):
        for split_name, data in results.items():
            _ap = credible.curves.average_metric(
                (data["precision"], data["recall"]),
            )
            ax.plot(
                data["curves"]["precision_recall"]["recall"],
                data["curves"]["precision_recall"]["precision"],
                label=f"{split_name} (AP: {_ap:.2f})",
            )
        ax.legend(loc="best", fancybox=True, framealpha=0.7)
        retval.append(fig)

    # score plots
    for split_name, data in results.items():
        score_fig = _score_plot(
            data["score-histograms"],
            f"Score distribution (split: {split_name})",
            data["threshold"],
        )
        retval.append(score_fig)

    return retval



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import functools
import pathlib

import click
from clapper.click import ResourceOption, verbosity_option

from ..utils.string import rewrap
from .click import ConfigCommand, PathOrURL
from .logging import setup_cli_logger

logger = setup_cli_logger()


def reusable_options(f):
    """Wrap reusable training script options (for ``experiment``).

    This decorator equips the target function ``f`` with all (reusable) ``train`` script
    options.

    Parameters
    ----------
    f
        The target function to equip with options.  This function must have parameters
        that accept such options.

    Returns
    -------
        The decorated version of function ``f``
    """

    @click.option(
        "--output-folder",
        "-o",
        help="Directory in which to store results (created if does not exist)",
        required=True,
        type=click.Path(
            file_okay=False,
            dir_okay=True,
            writable=True,
            path_type=pathlib.Path,
        ),
        default="results",
        show_default=True,
        cls=ResourceOption,
    )
    @click.option(
        "--model",
        "-m",
        help="A lightning module instance implementing the network to be trained",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--datamodule",
        "-d",
        help="A lightning DataModule containing the training and validation sets.",
        required=True,
        cls=ResourceOption,
    )
    @click.option(
        "--batch-size",
        "-b",
        help=rewrap(
            """Number of samples in every batch (this parameter affects memory
            requirements for the network).  If the number of samples in the batch is
            larger than the total number of samples available for training, this value
            is truncated.  If this number is smaller, then batches of the specified size
            are created and fed to the network until there are no more new samples to
            feed (epoch is finished). If the total number of training samples is not a
            multiple of the batch-size, the last batch will be smaller than the first,
            unless --drop-incomplete-batch is set, in which case this batch is not
            used."""
        ),
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--accumulate-grad-batches",
        "-a",
        help=rewrap(
            """Number of accumulations for backward propagation to accumulate "
            gradients over k batches before stepping the optimizer. This parameter, used
            in conjunction with the batch-size, may be used to reduce the number of
            samples loaded in each iteration, to affect memory usage in exchange for
            processing time (more iterations). This is useful interesting when one is
            training on GPUs with a limited amount of onboard RAM. The default of 1
            forces the whole batch to be processed at once. Otherwise the batch is
            multiplied by accumulate-grad-batches pieces, and gradients are accumulated
            to complete each training step."""
        ),
        required=True,
        show_default=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--drop-incomplete-batch/--no-drop-incomplete-batch",
        "-D",
        help=rewrap(
            """If set, the last batch in an epoch will be dropped if incomplete. If you
            set this option, you should also consider increasing the total number of
            epochs of training, as the total number of training steps may be
            reduced."""
        ),
        required=True,
        show_default=True,
        default=False,
        cls=ResourceOption,
    )
    @click.option(
        "--epochs",
        "-e",
        help=rewrap(
            """Number of epochs (complete training set passes) to train for. If
            continuing from a saved checkpoint, ensure to provide a greater number of
            epochs than was saved in the checkpoint to be loaded."""
        ),
        show_default=True,
        required=True,
        default=1000,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--validation-period",
        "-p",
        help=rewrap(
            """Number of epochs after which validation happens.  By default, we run
            validation after every training epoch (period=1).  You can change this to
            make validation more sparse, by increasing the validation period. Notice
            that this affects checkpoint saving.  While checkpoints are created after
            every training step (the last training step always triggers the overriding
            of latest checkpoint), and this process is independent of validation runs,
            evaluation of the 'best' model obtained so far based on those will be
            influenced by this setting."""
        ),
        show_default=True,
        required=True,
        default=1,
        type=click.IntRange(min=1),
        cls=ResourceOption,
    )
    @click.option(
        "--device",
        "-x",
        help=rewrap(
            """A string indicating the device to use (e.g. "cpu", "cuda", "mps",
            "cuda:0"). Can be passed as a string, or as an environment variable (e.g.
            `MEDNET_DEVICE=cpu` or `MEDNET_DEVICE=mps`)."""
        ),
        envvar="MEDNET_DEVICE",
        show_default=True,
        required=True,
        default="cpu",
        cls=ResourceOption,
    )
    @click.option(
        "--cache-samples/--no-cache-samples",
        help="""If set to True, loads the sample into memory, otherwise loads them at
        runtime.""",
        required=True,
        show_default=True,
        default=False,
        cls=ResourceOption,
    )
    @click.option(
        "--seed",
        "-s",
        help="Seed to use for the random number generator",
        show_default=True,
        required=False,
        default=42,
        type=click.IntRange(min=0),
        cls=ResourceOption,
    )
    @click.option(
        "--parallel",
        "-P",
        help=rewrap(
            """Use multiprocessing for data loading: if set to -1 (default), disables
            multiprocessing data loading.  Set to 0 to enable as many data loading
            instances as processing cores available in the system.  Set to >= 1 to
            enable that many multiprocessing instances for data loading."""
        ),
        type=click.IntRange(min=-1),
        show_default=True,
        required=True,
        default=-1,
        cls=ResourceOption,
    )
    @click.option(
        "--monitoring-interval",
        "-I",
        help=rewrap(
            """Time between checks for the use of resources during each training epoch,
            in seconds.  An interval of 5 seconds, for example, will lead to CPU and GPU
            resources being probed every 5 seconds during each training epoch. Values
            registered in the training logs correspond to averages (or maxima) observed
            through possibly many probes in each epoch. Notice that setting a very small
            value may cause the probing process to become extremely busy, potentially
            biasing the overall perception of resource usage."""
        ),
        type=click.FloatRange(min=0.1),
        show_default=True,
        required=True,
        default=5.0,
        cls=ResourceOption,
    )
    @click.option(
        "--augmentations",
        "-A",
        help=rewrap(
            """Models that can be trained in this package are shipped without explicit
            data augmentations. This option allows you to define a list of data
            augmentations to use for training the selected model."""
        ),
        type=click.UNPROCESSED,
        default=[],
        cls=ResourceOption,
    )
    @click.option(
        "--balance-classes/--no-balance-classes",
        "-B/-N",
        help=rewrap(
            """If set, balances the loss term to take into consideration the occurence
            of each target in the training and validation splits."""
        ),
        required=True,
        show_default=True,
        default=True,
        cls=ResourceOption,
    )
    @click.option(
        "--initial-weights",
        "-W",
        help=rewrap(
            """Path or URL to pretrained model file (`.ckpt` extension), corresponding
            to the architecture set with `--model`. Optionally, you may also pass a
            directory containing the result of a training session, in which case either
            the best (lowest validation) or latest model will be loaded. Note these
            weights will be overwritten when re-starting training from a checkpoint.
            This parameter should only be used as a set of initial weights (when
            starting a new training session)."""
        ),
        required=False,
        cls=ResourceOption,
        type=PathOrURL(
            exists=True,
            file_okay=True,
            dir_okay=True,
            readable=True,
            path_type=pathlib.Path,
        ),
    )
    @functools.wraps(f)
    def wrapper_reusable_options(*args, **kwargs):
        return f(*args, **kwargs)

    return wrapper_reusable_options


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Train a Pasa model with the montgomery dataset (classification task):

   .. code:: sh

      mednet train -vv pasa montgomery

2. Train a Little WNet model with the drive dataset (vessel segmentation task):

   .. code:: sh

      mednet train -vv lwnet drive

3. Train a Densenet model with the shenzhen dataset, while starting from a set
   of pre-trained weights:

   .. code:: sh

      mednet train -vv densenet shenzhen --initial-weights=path/to/checkpoint.ckpt
""",
)
@reusable_options
@verbosity_option(logger=logger, expose_value=False)
def train(
    model,
    output_folder,
    epochs,
    batch_size,
    accumulate_grad_batches,
    drop_incomplete_batch,
    datamodule,
    validation_period,
    device,
    cache_samples,
    seed,
    parallel,
    monitoring_interval,
    augmentations,
    balance_classes,
    initial_weights,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Train a model on a given datamodule (task-specific).

    Training is performed for a configurable number of epochs, and
    generates checkpoints.  Checkpoints are model files with a .ckpt
    extension that are used in subsequent tasks or from which training
    can be resumed.
    """
    from lightning.pytorch import seed_everything

    from ..engine.device import DeviceManager
    from ..engine.trainer import (
        get_checkpoint_file,
        load_checkpoint,
        run,
        setup_datamodule,
        validate_model_datamodule,
    )
    from ..models.model import Model
    from .utils import save_json_metadata

    validate_model_datamodule(model, datamodule)

    seed_everything(seed)

    # report model/transforms options - set data augmentations
    logger.info(f"Network model: {type(model).__module__}.{type(model).__name__}")
    model.augmentation_transforms = augmentations

    device_manager = DeviceManager(device)

    # reset datamodule with user configurable options
    setup_datamodule(
        datamodule, model, batch_size, drop_incomplete_batch, cache_samples, parallel
    )

    if initial_weights is not None:
        if isinstance(initial_weights, pathlib.Path):
            if initial_weights.is_dir():
                initial_weights = get_checkpoint_file(initial_weights)

        logger.info(f"Loading initial weights from `{initial_weights}`...")
        model = type(model).load_from_checkpoint(initial_weights, strict=False)

    # resets the number of outputs in a model, if necessary
    model.num_classes = datamodule.num_classes

    # If asked, rebalances the loss criterion based on the relative proportion
    # of class examples available in the training set.  Also affects the
    # validation loss if a validation set is available on the DataModule.
    if balance_classes:
        model.balance_losses(datamodule)

    checkpoint_file = get_checkpoint_file(output_folder)

    if checkpoint_file is None:
        if isinstance(model, Model):
            if not model.normalizer_is_set():
                model.set_normalizer_from_dataloader(
                    datamodule.unshuffled_train_dataloader()
                )
        else:
            from ..utils.string import rewrap

            logger.warning(
                rewrap(
                    f"""Model `{type(model)}` is not known. Skipping input normalization
                    from train dataloader setup (unsupported external model)."""
                ),
            )
    else:
        load_checkpoint(checkpoint_file)

    # stores all information we can think of, to reproduce this later
    save_json_metadata(
        output_file=output_folder / "train.meta.json",
        datamodule=datamodule,
        model=model,
        augmentations=augmentations,
        device_manager=device_manager,
        output_folder=output_folder,
        epochs=epochs,
        batch_size=batch_size,
        accumulate_grad_batches=accumulate_grad_batches,
        drop_incomplete_batch=drop_incomplete_batch,
        validation_period=validation_period,
        cache_samples=cache_samples,
        seed=seed,
        parallel=parallel,
        monitoring_interval=monitoring_interval,
        balance_classes=balance_classes,
        initial_weights=initial_weights,
    )

    logger.info(f"Training for at most {epochs} epochs.")

    run(
        model=model,
        datamodule=datamodule,
        validation_period=validation_period,
        device_manager=device_manager,
        max_epochs=epochs,
        output_folder=output_folder,
        monitoring_interval=monitoring_interval,
        accumulate_grad_batches=accumulate_grad_batches,
        checkpoint=checkpoint_file,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Montgomery DataModule for TB detection.

The standard digital image database for Tuberculosis was created by the
National Library of Medicine, Maryland, USA in collaboration with Shenzhen No.3
People’s Hospital, Guangdong Medical College, Shenzhen, China.

* Database references: :cite:p:`jaeger_two_2014`,

Data specifications:

* Raw data input (on disk):

  * PNG images 8 bit grayscale issued from digital radiography machines
  * Original resolution (height x width or width x height): 4020x4892 px or
    4892x4020 px
  * Samples: 138 images and associated labels

* Output image:

  * Transforms:

    * Load raw PNG with :py:mod:`PIL`
    * Remove black borders
    * Convert to torch tensor

  * Final specifications

    * Grayscale, encoded as a single plane tensor, 32-bit floats,
      square at most 4020 x 4020 pixels
    * Binary labels: 0 (healthy), 1 (active tuberculosis), encoded as a 1D
      torch float tensor.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..image_utils import remove_black_borders
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Montgomery dataset.

    Parameters
    ----------
    config_variable
        Key to search for in the configuration file for the root directory of
        this database.
    multiclass
        Set to ``True`` if the targets should be output as 2 distinct classes
        instead of a single (0/1) output.
    """

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    # config_variable: required so this loader can be used for the small
    # version of the Montgomery database as well.
    def __init__(
        self, config_variable: str = CONFIGURATION_KEY_DATADIR, multiclass: bool = False
    ):
        self.datadir = pathlib.Path(
            load_rc().get(config_variable, os.path.realpath(os.curdir)),
        )
        self.multiclass = multiclass

    def sample(self, sample: tuple[str, int, typing.Any | None]) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            Expects a tuple containing the path suffix, within the dataset root
            folder, where to find the image to be loaded, and an integer,
            representing the sample target.

        Returns
        -------
            The sample representation as a dictionary.
        """

        # N.B.: Montgomery images are encoded as grayscale PNGs, so no need to
        # convert them again with Image.convert("L").
        image = PIL.Image.open(self.datadir / sample[0])
        image, _ = remove_black_borders(image)
        image = to_dtype(to_image(image), torch.float32, scale=True)
        image = tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            1D torch float tensor.
        """
        if self.multiclass:
            if sample[1] == 0:
                return torch.FloatTensor([1, 0])

            return torch.FloatTensor([0, 1])

        return torch.FloatTensor([sample[1]])


class DataModule(CachingDataModule):
    """Montgomery DataModule for TB detection.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    multiclass
        Set to ``True`` if the targets should be output as 2 distinct classes
        instead of a single (0/1) output.
    """

    def __init__(
        self,
        split_path: pathlib.Path | importlib.resources.abc.Traversable,
        multiclass: bool = False,
    ):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(multiclass=multiclass),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`DenseNet-121 network architecture <densenet-pytorch_>`_, from :cite:p:`huang_densely_2017`."""

import logging
import typing

import torch
import torch.nn
import torch.optim.optimizer
import torch.utils.data
import torchvision.models as models

from ...data.typing import TransformSequence
from ..typing import Checkpoint
from .model import Model

logger = logging.getLogger(__name__)


class Densenet(Model):
    """`DenseNet-121 network architecture <densenet-pytorch_>`_, from :cite:p:`huang_densely_2017`.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    pretrained
        If set to True, loads pretrained model weights during initialization,
        else trains a new model.
    dropout
        Dropout rate after each dense layer.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = torch.nn.BCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
        dropout: float = 0.1,
        num_classes: int = 1,
    ):
        super().__init__(
            name="densenet",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        self.pretrained = pretrained
        self.dropout = dropout

        # Load pretrained model
        weights = None
        if self.pretrained:
            from ..normalizer import make_imagenet_normalizer

            Model.normalizer.fset(self, make_imagenet_normalizer())  # type: ignore[attr-defined]

            logger.info(f"Loading pretrained `{self.name}` model weights")
            weights = models.DenseNet121_Weights.DEFAULT

        self.model = models.densenet121(weights=weights, drop_rate=self.dropout)

        # output layer
        self.model.classifier = torch.nn.Linear(
            self.model.classifier.in_features, self.num_classes
        )

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        if self.num_classes != v:
            logger.info(
                f"Resetting `{self.name}` output classifier layer weights due "
                f"to a change in output size ({self.num_classes} -> {v})"
            )
            self.model.classifier = torch.nn.Linear(
                self.model.classifier.in_features, v
            )
            self._num_classes = v

    def on_load_checkpoint(self, checkpoint: Checkpoint) -> None:
        # support previous version of densenet (model_ft -> model)
        if any([k.startswith("model_ft") for k in checkpoint["state_dict"].keys()]):
            # convert all "model_ft" entries to "model"
            checkpoint["state_dict"] = {
                k.replace("model_ft", "model"): v
                for k, v in checkpoint["state_dict"].items()
            }

        # reset number of output classes if need be
        self.num_classes = checkpoint["state_dict"]["model.classifier.bias"].shape[0]

        # perform routine checkpoint loading
        super().on_load_checkpoint(checkpoint)

    def forward(self, x):
        x = self.normalizer(x)
        return self.model(x)



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Shenzhen DataModule for computer-aided semantic sementation of lungs.

The standard digital image database for Tuberculosis was created by the
National Library of Medicine, Maryland, USA in collaboration with Shenzhen No.3
People’s Hospital, Guangdong Medical College, Shenzhen, China. The Chest X-rays
are from out-patient clinics, and were captured as part of the daily routine
using Philips DR Digital Diagnose systems.

The database includes 336 cases with manifestation of tuberculosis, and 326
normal cases.  It contains a total  of 662 images.  Image size varies for each
X-ray. It is approximately 3K x 3K. One set of ground-truth lung annotations is
available for 566 of the 662 images.

* Database references:

  * Original data :cite:p:`jaeger_two_2014`
  * Splits: :cite:p:`gaal_attention_2020`

.. important:: **Raw data organization**

    The Shenzhen_ base datadir, which you should configure following the
    :ref:`mednet.setup` instructions, must contain at least these two
    subdirectories:

    - ``CXR_png/`` (directory containing the CXR images)
    - ``mask/`` (contains masks downloaded from `Shenzhen Annotations`_)

Data specifications:

* Raw data input (on disk):

  * PNG 8-bit RGB images issued from digital radiography machines (grayscale,
    but encoded as RGB images with "inverted" grayscale scale requiring special
    treatment).
  * Original resolution: variable width and height of 3000 x 3000 pixels or
    less
  * Samples: 566 images and associated labels

* Output image:

  * Transforms:

    * Load raw PNG with :py:mod:`PIL`
    * Torch center cropping to get square image

  * Final specifications:

    * Grayscale, encoded as a 3-plane plane tensor, 32-bit floats,
      square with varying resolutions, depending on the input image
    * Labels: Binary mask with annotated lungs (1 where lungs are; 0 otherwise)
    * Mask: Binary mask with all ones

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the shenzhen dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        # use image as a base since target() can be overriden by child class
        mask = torch.ones((1, image.shape[-2], image.shape[-1]), dtype=torch.float32)

        image = tv_tensors.Image(image)
        target = tv_tensors.Mask(target)
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """Shenzhen database for lung segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Padchest database for computer-aided diagnosis.

A large chest x-ray image dataset with multi-label annotated reports. This
dataset includes more than 160,000 images from 67,000 patients that were
interpreted and reported by radiologists at Hospital San Juan (Spain) from 2009
to 2017, covering six different position views and additional information on
image acquisition and patient demography.

We keep only postero-anterior "PA" images in our setup.

* Reference: :cite:p:`bustos_padchest_2020`
* Raw data input (on disk):

  * PNG grayscale 16-bit depth images
  * Resolution: varying resolution

* Labels: :cite:p:`bustos_padchest_2020`
* Output image:

  * Transforms:

    * Load raw 16-bit PNG with :py:mod:`PIL`
    * Remove excess black borders
    * Convert image to 32-bit floats between 0. and 1.
    * Convert to tensor

  * Final specifications

    * Grayscale, encoded as a 1-plane 32-bit float image, square with
      varying resolutions depending on the raw input image
    * Labels, in order (some of which may not be present in all splits):

      * COPD signs
      * Chilaiditi sign
      * NSG tube
      * abnormal foreign body
      * abscess
      * adenopathy
      * air bronchogram
      * air fluid level
      * air trapping
      * alveolar pattern
      * aortic aneurysm
      * aortic atheromatosis
      * aortic button enlargement
      * aortic elongation
      * aortic endoprosthesis
      * apical pleural thickening
      * artificial aortic heart valve
      * artificial heart valve
      * artificial mitral heart valve
      * asbestosis signs
      * ascendent aortic elongation
      * atelectasis
      * atelectasis basal
      * atypical pneumonia
      * axial hyperostosis
      * azygoesophageal recess shift
      * azygos lobe
      * blastic bone lesion
      * bone cement
      * bone metastasis
      * breast mass
      * bronchiectasis
      * bronchovascular markings
      * bullas
      * calcified adenopathy
      * calcified densities
      * calcified fibroadenoma
      * calcified granuloma
      * calcified mediastinal adenopathy
      * calcified pleural plaques
      * calcified pleural thickening
      * callus rib fracture
      * cardiomegaly
      * catheter
      * cavitation
      * central vascular redistribution
      * central venous catheter
      * central venous catheter via jugular vein
      * central venous catheter via subclavian vein
      * central venous catheter via umbilical vein
      * cervical rib
      * chest drain tube
      * chronic changes
      * clavicle fracture
      * consolidation
      * costochondral junction hypertrophy
      * costophrenic angle blunting
      * cyst
      * dai
      * descendent aortic elongation
      * dextrocardia
      * diaphragmatic eventration
      * double J stent
      * dual chamber device
      * electrical device
      * emphysema
      * empyema
      * end on vessel
      * endoprosthesis
      * endotracheal tube
      * esophagic dilatation
      * exclude
      * external foreign body
      * fibrotic band
      * fissure thickening
      * flattened diaphragm
      * fracture
      * gastrostomy tube
      * goiter
      * granuloma
      * ground glass pattern
      * gynecomastia
      * heart insufficiency
      * heart valve calcified
      * hemidiaphragm elevation
      * hiatal hernia
      * hilar congestion
      * hilar enlargement
      * humeral fracture
      * humeral prosthesis
      * hydropneumothorax
      * hyperinflated lung
      * hypoexpansion
      * hypoexpansion basal
      * increased density
      * infiltrates
      * interstitial pattern
      * kerley lines
      * kyphosis
      * laminar atelectasis
      * lepidic adenocarcinoma
      * lipomatosis
      * lobar atelectasis
      * loculated fissural effusion
      * loculated pleural effusion
      * lung metastasis
      * lung vascular paucity
      * lymphangitis carcinomatosa
      * lytic bone lesion
      * major fissure thickening
      * mammary prosthesis
      * mass
      * mastectomy
      * mediastinal enlargement
      * mediastinal mass
      * mediastinal shift
      * mediastinic lipomatosis
      * metal
      * miliary opacities
      * minor fissure thickening
      * multiple nodules
      * nephrostomy tube
      * nipple shadow
      * nodule
      * non axial articular degenerative changes
      * normal
      * obesity
      * osteopenia
      * osteoporosis
      * osteosynthesis material
      * pacemaker
      * pectum carinatum
      * pectum excavatum
      * pericardial effusion
      * pleural effusion
      * pleural mass
      * pleural plaques
      * pleural thickening
      * pneumomediastinum
      * pneumonia
      * pneumoperitoneo
      * pneumothorax
      * post radiotherapy changes
      * prosthesis
      * pseudonodule
      * pulmonary artery enlargement
      * pulmonary artery hypertension
      * pulmonary edema
      * pulmonary fibrosis
      * pulmonary hypertension
      * pulmonary mass
      * pulmonary venous hypertension
      * reservoir central venous catheter
      * respiratory distress
      * reticular interstitial pattern
      * reticulonodular interstitial pattern
      * rib fracture
      * right sided aortic arch
      * round atelectasis
      * sclerotic bone lesion
      * scoliosis
      * segmental atelectasis
      * single chamber device
      * soft tissue mass
      * sternoclavicular junction hypertrophy
      * sternotomy
      * subacromial space narrowing
      * subcutaneous emphysema
      * suboptimal study
      * superior mediastinal enlargement
      * supra aortic elongation
      * surgery
      * surgery breast
      * surgery heart
      * surgery humeral
      * surgery lung
      * surgery neck
      * suture material
      * thoracic cage deformation
      * total atelectasis
      * tracheal shift
      * tracheostomy tube
      * tuberculosis
      * tuberculosis sequelae
      * unchanged
      * vascular hilar enlargement
      * vascular redistribution
      * ventriculoperitoneal drain tube
      * vertebral anterior compression
      * vertebral compression
      * vertebral degenerative changes
      * vertebral fracture
      * volume loss

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources.abc
import os
import pathlib
import typing

import numpy
import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..image_utils import remove_black_borders
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name of this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the PadChest database."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(
                CONFIGURATION_KEY_DATADIR,
                os.path.realpath(os.curdir),
            ),
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the database root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The sample representation.
        """

        # N.B.: PadChest images are encoded as 16-bit grayscale images
        image = PIL.Image.open(self.datadir / sample[0])
        image, _ = remove_black_borders(image)
        image = numpy.array(image).astype(numpy.float32) / 65535
        image = to_dtype(to_image(image), torch.float32, scale=True)
        image = tv_tensors.Image(image)

        # use the code below to view generated images
        # from torchvision.transforms.v2.functional import to_pil_image
        # to_pil_image(tensor).show()
        # __import__("pdb").set_trace()

        return dict(image=image, target=self.target(sample), name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The labels corresponding to all radiological signs present in the
            specified sample, encapsulated as a 1D torch float tensor.
        """

        return torch.FloatTensor(sample[1])


class DataModule(CachingDataModule):
    """Padchest database for computer-aided diagnosis.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="classification",
            num_classes=193,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`Montgomery/Shenzhen/Indian aggregated database <mednet.data.classify.montgomery_shenzhen_indian>` (cross validation fold 7)."""

import importlib.resources

from mednet.data.classify.montgomery_shenzhen_indian import DataModule

datamodule = DataModule(
    "fold-7",
    (
        importlib.resources.files("mednet.config.classify.data.montgomery")
        / "fold-7.json",
        importlib.resources.files("mednet.config.classify.data.shenzhen")
        / "fold-7.json",
        importlib.resources.files("mednet.config.classify.data.indian") / "fold-7.json",
    ),
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pathlib

import click
from clapper.click import ConfigCommand, ResourceOption, verbosity_option

from ...logging import setup_cli_logger

logger = setup_cli_logger()


@click.command(
    entry_point_group="mednet.config",
    cls=ConfigCommand,
    epilog="""Examples:

1. Generate visualizations in the form of heatmaps from existing saliency maps for a dataset configuration:

   .. code:: sh

      mednet classify saliency view -vv pasa tbx11k-v1-healthy-vs-atb --input-folder=parent_folder/gradcam/ --output-folder=path/to/visualizations
""",
)
@click.option(
    "--model",
    "-m",
    help="A lightning module instance implementing the network to be used for applying the necessary data transformations.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--datamodule",
    "-d",
    help="A lightning DataModule containing the training, validation and test sets.",
    required=True,
    cls=ResourceOption,
)
@click.option(
    "--input-folder",
    "-i",
    help="Path to the directory containing the saliency maps for a specific visualization type.",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="visualizations",
    cls=ResourceOption,
)
@click.option(
    "--output-folder",
    "-o",
    help="Directory in which to store the visualizations (created if does not exist)",
    required=True,
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        path_type=pathlib.Path,
    ),
    default="visualizations",
    show_default=True,
    cls=ResourceOption,
)
@click.option(
    "--show-groundtruth/--no-show-groundtruth",
    "-G/-g",
    help="""If set, visualizations for ground truth labels will be generated.
    Only works for datasets with bounding boxes.""",
    is_flag=True,
    default=False,
    cls=ResourceOption,
)
@click.option(
    "--threshold",
    "-t",
    help="""The pixel values above ``threshold`` % of max value are kept in the
    original saliency map.  Everything else is set to zero.  The value proposed
    on :cite:p:`wang_score-cam_2020` is 0.2.  Use this value if unsure.""",
    show_default=True,
    required=True,
    default=0.2,
    type=click.FloatRange(min=0, max=1),
    cls=ResourceOption,
)
@verbosity_option(logger=logger, expose_value=False)
def view(
    model,
    datamodule,
    input_folder,
    output_folder,
    show_groundtruth,
    threshold,
    **_,
) -> None:  # numpydoc ignore=PR01
    """Generate heatmaps for input samples based on existing saliency maps."""

    from ....engine.classify.saliency.viewer import run
    from ...utils import save_json_metadata

    logger.info(f"Output folder: {output_folder}")
    output_folder.mkdir(parents=True, exist_ok=True)

    # register metadata
    save_json_metadata(
        output_file=output_folder / "view.meta.json",
        datamodule=datamodule,
        model=model,
        input_folder=input_folder,
        output_folder=output_folder,
        show_groundtruth=show_groundtruth,
        threshold=threshold,
    )

    datamodule.drop_incomplete_batch = False
    # datamodule.cache_samples = cache_samples
    # datamodule.parallel = parallel
    datamodule.model_transforms = model.model_transforms

    datamodule.prepare_data()
    datamodule.setup(stage="predict")

    run(
        datamodule=datamodule,
        input_folder=input_folder,
        target_label=1,
        output_folder=output_folder,
        show_groundtruth=show_groundtruth,
        threshold=threshold,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""`Little W-Net (LWNET) network architecture <lwnet_>`_, from :cite:p:`galdran_state---art_2022`.

It is based on two simple U-Nets with 3 layers concatenated to each other.  The first
U-Net produces a segmentation map that is used by the second to better guide
segmentation.
"""

import logging
import typing

import torch
import torch.nn

from ...data.typing import TransformSequence
from ...utils.string import rewrap
from ..losses import MultiLayerBCELogitsLossWeightedPerBatch
from .model import Model

logger = logging.getLogger(__name__)


def _conv1x1(in_planes, out_planes, stride=1):
    return torch.nn.Conv2d(
        in_planes, out_planes, kernel_size=1, stride=stride, bias=False
    )


class ConvBlock(torch.nn.Module):
    """Convolution block.

    Parameters
    ----------
    input_channels
        Number of input channels.
    output_channels
        Number of output channels.
    kernel_size
        Kernel Size.
    add_conv2d
        If True, adds a Conv2d layer.
    add_max_pool2d
        If True, adds a MaxPool2d layer.
    """

    def __init__(
        self,
        input_channels,
        output_channels,
        kernel_size=3,
        add_conv2d=False,
        add_max_pool2d=True,
    ):
        super().__init__()
        if add_conv2d is True:
            self.conv2d_layer = torch.nn.Sequential(
                _conv1x1(input_channels, output_channels),
                torch.nn.BatchNorm2d(output_channels),
            )
        else:
            self.conv2d_layer = None
        pad = (kernel_size - 1) // 2

        block = []
        self.pool = None
        if add_max_pool2d:
            self.pool = torch.nn.MaxPool2d(kernel_size=2)

        block.append(
            torch.nn.Conv2d(
                input_channels, output_channels, kernel_size=kernel_size, padding=pad
            )
        )
        block.append(torch.nn.ReLU())
        block.append(torch.nn.BatchNorm2d(output_channels))

        block.append(
            torch.nn.Conv2d(
                output_channels, output_channels, kernel_size=kernel_size, padding=pad
            )
        )
        block.append(torch.nn.ReLU())
        block.append(torch.nn.BatchNorm2d(output_channels))

        self.block = torch.nn.Sequential(*block)

    def forward(self, x: torch.Tensor):
        if self.pool is not None:
            x = self.pool(x)
        out = self.block(x)
        if self.conv2d_layer is not None:
            return out + self.conv2d_layer(x)
        return out


class UpsampleBlock(torch.nn.Module):
    """Upsample block implementation.

    Parameters
    ----------
    input_channels
        Number of input channels.
    output_channels
        Number of output channels.
    upsampling_mode
        Upsampling mode.
    """

    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        upsampling_mode: typing.Literal["transp_conv", "up_conv"],
    ):
        super().__init__()
        block = []

        match upsampling_mode:
            case "transp_conv":
                block.append(
                    torch.nn.ConvTranspose2d(
                        input_channels, output_channels, kernel_size=2, stride=2
                    )
                )
            case "up_conv":
                block.append(
                    torch.nn.Upsample(
                        mode="bilinear", scale_factor=2, align_corners=False
                    )
                )
                block.append(
                    torch.nn.Conv2d(input_channels, output_channels, kernel_size=1)
                )
            case _:
                raise NotImplementedError(
                    "Upsampling mode `{upsampling_mode}` is not supported"
                )

        self.block = torch.nn.Sequential(*block)

    def forward(self, x):
        return self.block(x)


class ConvBridgeBlock(torch.nn.Module):
    """ConvBridgeBlock implementation.

    Parameters
    ----------
    channels
        Number of channels.
    kernel_size
        Kernel Size.
    """

    def __init__(self, channels: int, kernel_size: int):
        super().__init__()
        pad = (kernel_size - 1) // 2
        block = []

        block.append(
            torch.nn.Conv2d(channels, channels, kernel_size=kernel_size, padding=pad)
        )
        block.append(torch.nn.ReLU())
        block.append(torch.nn.BatchNorm2d(channels))

        self.block = torch.nn.Sequential(*block)

    def forward(self, x):
        return self.block(x)


class UpConvBlock(torch.nn.Module):
    """UpConvBlock implementation.

    Parameters
    ----------
    input_channels
        Number of input channels.
    output_channels
        Number of output channels.
    kernel_size
        Kernel Size.
    upsampling_mode
        Upsampling mode.
    add_conv_bridge
        If True, adds a ConvBridgeBlock layer.
    add_conv2d
        If True, adds a Conv2d layer.
    """

    def __init__(
        self,
        input_channels: int,
        output_channels: int,
        kernel_size: int = 3,
        upsampling_mode: typing.Literal["transp_conv", "up_conv"] = "up_conv",
        add_conv_bridge: bool = False,
        add_conv2d: bool = False,
    ):
        super().__init__()
        self.add_conv_bridge = add_conv_bridge

        self.up_layer = UpsampleBlock(
            input_channels=input_channels,
            output_channels=output_channels,
            upsampling_mode=upsampling_mode,
        )
        self.conv_layer = ConvBlock(
            2 * output_channels,
            output_channels,
            kernel_size=kernel_size,
            add_conv2d=add_conv2d,
            add_max_pool2d=False,
        )
        if self.add_conv_bridge:
            self.add_conv_bridge_layer = ConvBridgeBlock(
                channels=output_channels, kernel_size=kernel_size
            )

    def forward(self, x, skip):
        up = self.up_layer(x)
        if self.add_conv_bridge:
            out = torch.cat([up, self.add_conv_bridge_layer(skip)], dim=1)
        else:
            out = torch.cat([up, skip], dim=1)
        return self.conv_layer(out)


class LittleUNet(torch.nn.Module):
    """Base little U-Net (LUNET) network architecture, from :cite:p:`galdran_state---art_2022`.

    Parameters
    ----------
    input_channels
        Number of input channels.
    layers
        Number of layers of the model.
    num_classes
        Number of outputs (classes) for this model.
    kernel_size
        Kernel Size.
    upsampling_mode
        Upsampling mode.
    add_conv_bridge
        If True, adds a ConvBridgeBlock layer.
    add_conv2d
        If True, adds a Conv2d layer.
    """

    def __init__(
        self,
        input_channels,
        layers,
        num_classes: int = 1,
        kernel_size: int = 3,
        upsampling_mode: typing.Literal["transp_conv", "up_conv"] = "transp_conv",
        add_conv_bridge=True,
        add_conv2d=True,
    ):
        super().__init__()

        self.name = "lunet"
        self._num_classes = num_classes
        self.layers = layers
        self.first = ConvBlock(
            input_channels=input_channels,
            output_channels=self.layers[0],
            kernel_size=kernel_size,
            add_conv2d=add_conv2d,
            add_max_pool2d=False,
        )

        self.down_path = torch.nn.ModuleList()
        for i in range(len(self.layers) - 1):
            block = ConvBlock(
                input_channels=self.layers[i],
                output_channels=self.layers[i + 1],
                kernel_size=kernel_size,
                add_conv2d=add_conv2d,
                add_max_pool2d=True,
            )
            self.down_path.append(block)

        self.up_path = torch.nn.ModuleList()
        reversed_layers = list(reversed(self.layers))
        for i in range(len(self.layers) - 1):
            block = UpConvBlock(
                input_channels=reversed_layers[i],
                output_channels=reversed_layers[i + 1],
                kernel_size=kernel_size,
                upsampling_mode=upsampling_mode,
                add_conv_bridge=add_conv_bridge,
                add_conv2d=add_conv2d,
            )
            self.up_path.append(block)

        # init, shamelessly lifted from torchvision/models/resnet.py
        for m in self.modules():
            if isinstance(m, torch.nn.Conv2d):
                torch.nn.init.kaiming_normal_(
                    m.weight, mode="fan_out", nonlinearity="relu"
                )
            elif isinstance(m, torch.nn.BatchNorm2d | torch.nn.GroupNorm):
                torch.nn.init.constant_(m.weight, 1)
                torch.nn.init.constant_(m.bias, 0)

        self.final = torch.nn.Conv2d(self.layers[0], self._num_classes, kernel_size=1)

    def forward(self, x):
        x = self.first(x)
        down_activations = []
        for i, down in enumerate(self.down_path):
            down_activations.append(x)
            x = down(x)
        down_activations.reverse()
        for i, up in enumerate(self.up_path):
            x = up(x, down_activations[i])
        return self.final(x)

    @property
    def num_classes(self) -> int:
        """Number of outputs (classes) for this model.

        Returns
        -------
        int
            The number of outputs supported by this model.
        """
        return self._num_classes

    @num_classes.setter
    def num_classes(self, v: int) -> None:
        if self._num_classes != v:
            logger.info(
                rewrap(
                    f"""Resetting `{self.name}` unet1 conv2d layer weights and the
                    **whole** unet2 due to a change in output size ({self.num_classes}
                    -> {v})"""
                )
            )
            self.final = torch.nn.Conv2d(self.layers[0], v, kernel_size=1)
            self._num_classes = v


class LittleWNet(Model):
    """`Little W-Net (LWNET) network architecture <lwnet_>`_, from :cite:p:`galdran_state---art_2022`.

    Concatenates two :py:class:`Little U-Net <LittleUNet>` models.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed
           to the batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be
        applied on the input **before** it is fed into the network.
    num_classes
        Number of outputs (classes) for this model.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = MultiLayerBCELogitsLossWeightedPerBatch,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        num_classes: int = 1,
    ):
        super().__init__(
            name="lwnet",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=num_classes,
        )

        self.unet1 = LittleUNet(
            input_channels=3,
            num_classes=self.num_classes,
            layers=(8, 16, 32),
            add_conv_bridge=True,
            add_conv2d=True,
        )
        self.unet2 = LittleUNet(
            input_channels=3 + self.num_classes,
            num_classes=self.num_classes,
            layers=(8, 16, 32),
            add_conv_bridge=True,
            add_conv2d=True,
        )

    @Model.num_classes.setter  # type: ignore[attr-defined]
    def num_classes(self, v: int) -> None:
        if self.num_classes != v:
            logger.info(
                rewrap(
                    f"""Resetting `{self.name}` unet1 conv2d layer weights and the
                    **whole** unet2 due to a change in output size ({self.num_classes}
                    -> {v})"""
                )
            )
            self.unet1.num_classes = v
            self.unet2 = LittleUNet(
                input_channels=3 + v,
                num_classes=v,
                layers=(8, 16, 32),
                add_conv_bridge=True,
                add_conv2d=True,
            )
            self._num_classes = v

    def forward(self, x):
        x = self.normalizer(x)
        x1 = self.unet1(x)
        x2 = self.unet2(torch.cat([x, x1], dim=1))
        return x1, x2

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        del batch_idx, dataloader_idx  # satisfies linter
        # prediction only returns the result of the last lunet
        return torch.sigmoid(self(batch["image"])[1])



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""CHASE-DB1 dataset for vessel segmentation.

The CHASE_DB1 is a retinal vessel reference dataset acquired from multiethnic
school children. This database is a part of the Child Heart and Health Study in
England (CHASE), a cardiovascular health survey in 200 primary schools in
London, Birmingham, and Leicester. The ocular imaging was carried out in 46
schools and demonstrated associations between retinal vessel tortuosity and
early risk factors for cardiovascular disease in over 1000 British primary
school children of different ethnic origin. The retinal images of both of the
eyes of each child were recorded with a hand-held Nidek NM-200-D fundus camera.
The images were captured at 30 degrees FOV camera. The dataset of images are
characterized by having nonuniform back-ground illumination, poor contrast of
blood vessels as compared with the background and wider arteriolars that have a
bright strip running down the centre known as the central vessel reflex.

* Reference: :cite:p:`fraz_ensemble_2012`

Data specifications:

* Raw data input (on disk):

  * RGB images encoded in JPG format with resolution (HxW) = 960 x 999 pixels.
  * Vessel annotations are encoded as PNG images with the same resolution as
    input samples.
  * Masks for the eye fundus are provided by this package.
  * Total samples: 28

* Output sample:

    * Image: Load raw JPG images with :py:mod:`PIL`, with auto-conversion to RGB.
    * Vessel annotations: Load annotations with :py:mod:`PIL`, with
      auto-conversion to model ``1`` with no dithering.
    * Eye fundus mask: Load mask with :py:mod:`PIL`, with
      auto-conversion to model ``1`` with no dithering.

Split ``first-annotator`` contains 8 training samples and 20 tests samples
annotated by expert 1.  Split ``second-annotator`` contains the sample samples
as in ``first-annotator``, but annotated by expert 2.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Chase-db1 dataset."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing path suffixes to the sample image, target, and mask
            to be loaded, within the dataset root folder.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = PIL.Image.open(path).convert(mode="1", dither=None)
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """CHASE-DB1 dataset for vessel segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""STARE dataset for vessel segmentation.

A subset of the original STARE dataset contains 20 annotated eye fundus images
with a resolution of 700 x 605 (width x height). Two sets of ground-truth
vessel annotations are available. The first set by Adam Hoover ("ah") is
commonly used for training and testing. The second set by Valentina Kouznetsova
("vk") is typically used as a “human” baseline.

* Database references:

  * Original data: :cite:p:`hoover_locating_2000`
  * Split reference: :cite:p:`maninis_deep_2016`

Data specifications:

* Raw data input (on disk):

  * RGB images encoded in PPM format with resolution (HxW) = 605 x 700
  * Total samples: 397 (out of which only 20 are annotated for vessel
    segmentation)

* Output sample:

    * Image: Load raw PPM images with :py:mod:`PIL`, with auto-conversion to RGB.
    * Vessel annotations: Load annotations with :py:mod:`PIL`, with
      auto-conversion to model ``1`` with no dithering.
    * Eye fundus mask: Load mask with :py:mod:`PIL`, with
      auto-conversion to model ``1`` with no dithering.

Protocol ``ah`` (default baseline, with first, more detailed annotator)
includes 10 training samples and 10 test samples.  Protocol ``vk`` (second
annotator) includes the same samples but annotated by a second expert.

This module contains the base declaration of common data modules and raw-data
loaders for this database. All configured splits inherit from this definition.
"""

import importlib.resources
import importlib.resources.abc
import os
import pathlib
import typing

import PIL.Image
import torch
from torchvision import tv_tensors
from torchvision.transforms.v2.functional import to_dtype, to_image

from ...models.transforms import crop_image_to_mask
from ...utils.rc import load_rc
from ..datamodule import CachingDataModule
from ..split import JSONDatabaseSplit
from ..typing import RawDataLoader as BaseDataLoader
from ..typing import Sample

DATABASE_SLUG = __name__.rsplit(".", 1)[-1]
"""Pythonic name to refer to this database."""

CONFIGURATION_KEY_DATADIR = "datadir." + DATABASE_SLUG
"""Key to search for in the configuration file for the root directory of this
database."""


class RawDataLoader(BaseDataLoader):
    """A specialized raw-data-loader for the Stare database."""

    datadir: pathlib.Path
    """This variable contains the base directory where the database raw data is
    stored."""

    def __init__(self):
        self.datadir = pathlib.Path(
            load_rc().get(CONFIGURATION_KEY_DATADIR, os.path.realpath(os.curdir))
        )

    def sample(self, sample: typing.Any) -> Sample:
        """Load a single image sample from the disk.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the database root folder,
            where to find the image to be loaded, and an integer, representing the
            sample label.

        Returns
        -------
            The sample representation.
        """

        image = PIL.Image.open(self.datadir / sample[0]).convert(mode="RGB")
        image = to_dtype(to_image(image), torch.float32, scale=True)

        target = self.target(sample)

        mask_path = (
            importlib.resources.files(__package__) / "masks" / DATABASE_SLUG / sample[2]
        )
        with importlib.resources.as_file(mask_path) as path:
            mask = PIL.Image.open(path).convert(mode="1", dither=None)
            mask = to_dtype(to_image(mask), torch.float32, scale=True)

        image = tv_tensors.Image(crop_image_to_mask(image, mask))
        target = tv_tensors.Mask(crop_image_to_mask(target, mask))
        mask = tv_tensors.Mask(mask)

        return dict(image=image, target=target, mask=mask, name=sample[0])

    def target(self, sample: typing.Any) -> torch.Tensor:
        """Load only sample target from its raw representation.

        Parameters
        ----------
        sample
            A tuple containing the path suffix, within the dataset root folder,
            where to find the image to be loaded, and an integer, representing
            the sample target.

        Returns
        -------
            The label corresponding to the specified sample, encapsulated as a
            torch float tensor.
        """

        target = PIL.Image.open(self.datadir / sample[1]).convert(mode="1", dither=None)
        return to_dtype(to_image(target), torch.float32, scale=True)


class DataModule(CachingDataModule):
    """STARE database for Vessel Segmentation.

    Parameters
    ----------
    split_path
        Path or traversable (resource) with the JSON split description to load.
    """

    def __init__(self, split_path: pathlib.Path | importlib.resources.abc.Traversable):
        super().__init__(
            database_split=JSONDatabaseSplit(split_path),
            raw_data_loader=RawDataLoader(),
            database_name=DATABASE_SLUG,
            split_name=split_path.name.rsplit(".", 2)[0],
            task="segmentation",
            num_classes=1,
        )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import random

import numpy

from mednet.engine.segment.evaluator import all_metrics


def test_all_metrics():
    tp = random.randint(1, 100)
    fp = random.randint(1, 100)
    tn = random.randint(1, 100)
    fn = random.randint(1, 100)

    p, r, s, a, j, f1 = all_metrics(tp, fp, tn, fn)

    assert (tp / (tp + fp)) == p
    assert (tp / (tp + fn)) == r
    assert (tn / (tn + fp)) == s
    assert ((tp + tn) / (tp + tn + fp + fn)) == a
    assert (tp / (tp + fp + fn)) == j
    assert ((2.0 * tp) / (2.0 * tp + fp + fn)) == f1
    assert numpy.isclose((2 * p * r) / (p + r), f1)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for transforms."""

import numpy
import PIL.Image
import torch
import torchvision.transforms.v2.functional as F  # noqa: N812
import torchvision.tv_tensors

import mednet.data.augmentations
import mednet.models.transforms


def test_crop_mask():
    original_tensor_size = (3, 50, 100)
    original_mask_size = (1, 50, 100)
    slice_ = (slice(None), slice(10, 30), slice(50, 70))

    tensor = torch.rand(original_tensor_size)
    mask = torch.zeros(original_mask_size)
    mask[slice_] = 1

    cropped_tensor = mednet.models.transforms.crop_image_to_mask(tensor, mask)

    assert cropped_tensor.shape == (3, 20, 20)
    assert torch.all(cropped_tensor.eq(tensor[slice_]))


def test_elastic_deformation(datadir):
    # Get a raw sample without deformation
    raw_without_deformation = PIL.Image.open(
        datadir / "raw_without_elastic_deformation.png"
    )
    raw_without_deformation = F.to_dtype(
        F.to_image(raw_without_deformation), torch.float32, scale=True
    )

    # Elastic deforms the raw
    ed = mednet.data.augmentations.ElasticDeformation(seed=100)
    raw_deformed = ed(raw_without_deformation)
    raw_deformed = F.to_pil_image(raw_deformed)
    # uncomment to save a new reference if required
    # raw_deformed.save(datadir / "raw_with_elastic_deformation.png")

    # Get the same sample already deformed (with seed=100)
    raw_2 = PIL.Image.open(datadir / "raw_with_elastic_deformation.png")
    raw_2 = F.to_pil_image(F.to_dtype(F.to_image(raw_2), torch.float32, scale=True))

    numpy.testing.assert_array_equal(numpy.asarray(raw_deformed), numpy.asarray(raw_2))


def test_square_center_pad():
    # is this transform functionally correct?

    tensor = torch.rand((3, 50, 100)).float()
    x = mednet.models.transforms.square_center_pad(tensor, tensor.shape[-2:])
    assert x.shape == (3, 100, 100)
    assert (tensor == x[:, 25:75, :]).all()
    assert (x[:, 0:25, :] == 0.0).all()
    assert (x[:, 75:, :] == 0.0).all()

    tensor = torch.rand((3, 12, 10)).float()
    x = mednet.models.transforms.square_center_pad(tensor, tensor.shape[-2:])
    assert x.shape == (3, 12, 12)
    assert (tensor == x[:, :, 1:11]).all()
    assert (x[:, :, 0] == 0.0).all()
    assert (x[:, :, 11] == 0.0).all()


def test_square_center_pad_tv_v2():
    # does this transform conform to the torchvision v2 transform API?

    tensor = torch.rand((3, 50, 100)).float()

    transform = mednet.models.transforms.SquareCenterPad()
    x = transform(tensor)

    assert isinstance(x, torch.Tensor)
    assert not isinstance(x, torchvision.tv_tensors.Image)
    assert x.shape == (3, 100, 100)
    assert (tensor == x[:, 25:75, :]).all()
    assert (x[:, 0:25, :] == 0.0).all()
    assert (x[:, 75:, :] == 0.0).all()

    image = torchvision.tv_tensors.Image(tensor)

    xi = transform(image)
    assert isinstance(xi, torchvision.tv_tensors.Image)
    assert xi.shape == (3, 100, 100)
    assert (tensor == xi[:, 25:75, :]).all()
    assert (xi[:, 0:25, :] == 0.0).all()
    assert (xi[:, 75:, :] == 0.0).all()

    mask = torchvision.tv_tensors.Mask(tensor)

    xm = transform(mask)
    assert isinstance(xm, torchvision.tv_tensors.Mask)
    assert xi.shape == (3, 100, 100)
    assert (tensor == xi[:, 25:75, :]).all()
    assert (xi[:, 0:25, :] == 0.0).all()
    assert (xi[:, 75:, :] == 0.0).all()

    bboxes = torchvision.tv_tensors.BoundingBoxes(
        [[10, 10, 30, 30]], format="XYXY", canvas_size=tensor.shape[-2:]
    )

    xb = transform(bboxes)
    assert isinstance(xb, torchvision.tv_tensors.BoundingBoxes)
    assert xb.canvas_size == (100, 100)
    assert (xb == torch.Tensor([[10, 10 + 25, 30, 30 + 25]])).all()



# SPDX-FileCopyrightText: Copyright © 2024 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
""":py:mod:`RIM-ONE r3 database <mednet.data.segment.rimoner3>` (default split for optic-disc segmentation with expert 2 annotations)."""

import importlib.resources

from mednet.data.segment.rimoner3 import DataModule

datamodule = DataModule(
    importlib.resources.files(__package__ or __name__.rsplit(".", 1)[0])
    / "optic-disc-exp2.json"
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Functions supporting training of pytorch models."""

import logging
import os
import pathlib

import lightning.pytorch
import lightning.pytorch.callbacks
import lightning.pytorch.loggers
import torch
import torch.utils.data

from ..data.datamodule import ConcatDataModule
from ..models.model import Model
from ..utils.checkpointer import CHECKPOINT_ALIASES
from ..utils.resources import ResourceMonitor
from .callbacks import LoggingCallback
from .device import DeviceManager

logger = logging.getLogger(__name__)


def get_checkpoint_file(results_dir: pathlib.Path) -> pathlib.Path | None:
    """Return the path of the latest checkpoint if it exists.

    Parameters
    ----------
    results_dir
        Directory in which results are saved.

    Returns
    -------
        Path to the latest checkpoint
    """
    from ..utils.checkpointer import get_checkpoint_to_resume_training

    checkpoint_file = None
    if results_dir.is_dir():
        try:
            checkpoint_file = get_checkpoint_to_resume_training(results_dir)
        except FileNotFoundError:
            logger.info(
                f"Folder {results_dir} already exists, but I did not"
                f" find any usable checkpoint file to resume training"
                f" from. Starting from scratch...",
            )

    return checkpoint_file


def load_checkpoint(checkpoint_file: pathlib.Path):
    """Load the checkpoint.

    Parameters
    ----------
    checkpoint_file
        Path to the checkpoint.
    """

    # Normalizer will be loaded during model.on_load_checkpoint
    checkpoint = torch.load(checkpoint_file)
    start_epoch = checkpoint["epoch"]
    logger.info(
        f"Resuming from epoch {start_epoch} "
        f"(checkpoint file: `{str(checkpoint_file)}`)...",
    )


def setup_datamodule(
    datamodule: ConcatDataModule,
    model: Model,
    batch_size: int,
    drop_incomplete_batch: bool,
    cache_samples: bool,
    parallel: int,
) -> None:  # numpydoc ignore=PR01
    """Configure and set up the datamodule."""
    datamodule.batch_size = batch_size
    datamodule.drop_incomplete_batch = drop_incomplete_batch
    datamodule.cache_samples = cache_samples
    datamodule.parallel = parallel
    datamodule.model_transforms = model.model_transforms

    datamodule.prepare_data()
    datamodule.setup(stage="fit")


def validate_model_datamodule(model: Model, datamodule: ConcatDataModule):
    """Validate the use of a model and datamodule together.

    Parameters
    ----------
    model
        The model to be validated.
    datamodule
        The datamodule to be validated.

    Raises
    ------
    TypeError
        In case the types of both objects is not compatible.
    """

    from ..models.classify.model import Model as ClassificationModel
    from ..models.detect.model import Model as DetectionModel
    from ..models.segment.model import Model as SegmentationModel

    # asserts data module and model are compatible
    match model:
        case ClassificationModel():
            if datamodule.task != "classification":
                raise TypeError(
                    f"Classification model `{model.name}` is incompatible with "
                    f"`{datamodule.task}` task from datamodule "
                    f"`{datamodule.database_name}`."
                )
        case SegmentationModel():
            if datamodule.task != "segmentation":
                raise TypeError(
                    f"Segmentation model `{model.name}` is incompatible with "
                    f"`{datamodule.task}` task from datamodule "
                    f"`{datamodule.database_name}`."
                )
        case DetectionModel():
            if datamodule.task != "detection":
                raise TypeError(
                    f"Detection model `{model.name}` is incompatible with "
                    f"`{datamodule.task}` task from datamodule "
                    f"`{datamodule.database_name}`."
                )
        case _:
            raise TypeError(f"Do not know how to handle model of type `{type(model)}`")


def run(
    model: lightning.pytorch.LightningModule,
    datamodule: lightning.pytorch.LightningDataModule,
    validation_period: int,
    device_manager: DeviceManager,
    max_epochs: int,
    output_folder: pathlib.Path,
    monitoring_interval: int | float,
    accumulate_grad_batches: int,
    checkpoint: pathlib.Path | None,
):
    """Fit a CNN model using supervised learning and save it to disk.

    This method supports periodic checkpointing and the output of a
    tensorboard-formatted log with the evolution of some figures during training.

    Parameters
    ----------
    model
        Neural network model (e.g. pasa).
    datamodule
        The lightning DataModule to use for training **and** validation.
    validation_period
        Number of epochs after which validation happens.  By default, we run
        validation after every training epoch (period=1).  You can change this
        to make validation more sparse, by increasing the validation period.
        Notice that this affects checkpoint saving.  While checkpoints are
        created after every training step (the last training step always
        triggers the overriding of latest checkpoint), and that this process is
        independent of validation runs, evaluation of the 'best' model obtained
        so far based on those will be influenced by this setting.
    device_manager
        An internal device representation, to be used for training and
        validation.  This representation can be converted into a pytorch device
        or a lightning accelerator setup.
    max_epochs
        The maximum number of epochs to train for.
    output_folder
        Folder in which the results will be saved.
    monitoring_interval
        Interval, in seconds (or fractions), through which we should monitor
        resources during training.
    accumulate_grad_batches
        Number of accumulations for backward propagation to accumulate gradients
        over k batches before stepping the optimizer. The default of 1 forces
        the whole batch to be processed at once. Otherwise the batch is multiplied
        by accumulate-grad-batches pieces, and gradients are accumulated to complete
        each step. This is especially interesting when one is training on GPUs with
        a limited amount of onboard RAM.
    checkpoint
        Path to an optional checkpoint file to load.
    """

    output_folder.mkdir(parents=True, exist_ok=True)

    from .loggers import CustomTensorboardLogger

    log_dir = "logs"
    tensorboard_logger = CustomTensorboardLogger(
        output_folder,
        log_dir,
    )
    logger.info(
        f"Monitor training with `tensorboard serve "
        f"--logdir={output_folder}/{log_dir}/`. "
        f"Then, open a browser on the printed address.",
    )

    resource_monitor = ResourceMonitor(
        interval=monitoring_interval,
        device_type=device_manager.device_type,
        main_pid=os.getpid(),
    )

    # This checkpointer will operate at the end of every validation epoch
    # (which happens at each checkpoint period), it will then save the lowest
    # validation loss model observed.  It will also save the last trained model
    checkpoint_minvalloss_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        dirpath=output_folder,
        filename=CHECKPOINT_ALIASES["best"],
        save_last=True,  # will (re)create the last trained model, at every iteration
        monitor="loss/validation",
        mode="min",
        save_on_train_epoch_end=True,
        every_n_epochs=validation_period,  # frequency at which it checks the "monitor"
        enable_version_counter=False,  # no versioning of aliased checkpoints
    )
    checkpoint_minvalloss_callback.CHECKPOINT_NAME_LAST = CHECKPOINT_ALIASES[  # type: ignore
        "periodic"
    ]

    # This context guarantees that the start/stop of the monitoring thread will work
    # irrespectively of exceptions thrown by the code
    with resource_monitor:
        accelerator, devices = device_manager.lightning_accelerator()
        trainer = lightning.pytorch.Trainer(
            accelerator=accelerator,
            devices=devices,
            max_epochs=max_epochs,
            accumulate_grad_batches=accumulate_grad_batches,
            logger=tensorboard_logger,
            check_val_every_n_epoch=validation_period,
            log_every_n_steps=len(datamodule.train_dataloader()),
            callbacks=[
                LoggingCallback(resource_monitor),
                checkpoint_minvalloss_callback,
            ],
        )

        checkpoint_str = checkpoint if checkpoint is None else str(checkpoint)
        _ = trainer.fit(model, datamodule, ckpt_path=checkpoint_str)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""DRIU network architecture for optic-disc segmentation, from :cite:p:`maninis_deep_2016`."""

import logging
import typing

import torch
import torch.nn
import torch.utils.data

from ...data.typing import TransformSequence
from ..losses import SoftJaccardAndBCEWithLogitsLoss
from .backbones.vgg import vgg16_for_segmentation
from .driu import ConcatFuseBlock
from .make_layers import UpsampleCropBlock
from .model import Model

logger = logging.getLogger(__name__)


class DRIUODHead(torch.nn.Module):
    """DRIU for optic disc segmentation head module.

    Parameters
    ----------
    in_channels_list : list
        Number of channels for each feature map that is returned from backbone.
    """

    def __init__(self, in_channels_list):
        super().__init__()
        (
            in_upsample2,
            in_upsample_4,
            in_upsample_8,
            in_upsample_16,
        ) = in_channels_list

        self.upsample2 = UpsampleCropBlock(in_upsample2, 16, 4, 2, 0)
        # Upsample layers
        self.upsample4 = UpsampleCropBlock(in_upsample_4, 16, 8, 4, 0)
        self.upsample8 = UpsampleCropBlock(in_upsample_8, 16, 16, 8, 0)
        self.upsample16 = UpsampleCropBlock(in_upsample_16, 16, 32, 16, 0)

        # Concat and Fuse
        self.concatfuse = ConcatFuseBlock()

    def forward(self, x):
        hw = x[0]
        upsample2 = self.upsample2(x[1], hw)  # side-multi2-up
        upsample4 = self.upsample4(x[2], hw)  # side-multi3-up
        upsample8 = self.upsample8(x[3], hw)  # side-multi4-up
        upsample16 = self.upsample16(x[4], hw)  # side-multi5-up
        return self.concatfuse(upsample2, upsample4, upsample8, upsample16)


class DRIUOD(Model):
    """DRIU network architecture for optic-disc segmentation, from :cite:p:`maninis_deep_2016`.

    Parameters
    ----------
    loss_type
        The loss to be used for training and evaluation.

        .. warning::

           The loss should be set to always return batch averages (as opposed to the
           batch sum), as our logging system expects it so.
    loss_arguments
        Arguments to the loss.
    optimizer_type
        The type of optimizer to use for training.
    optimizer_arguments
        Arguments to the optimizer after ``params``.
    scheduler_type
        The type of scheduler to use for training.
    scheduler_arguments
        Arguments to the scheduler after ``params``.
    model_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    augmentation_transforms
        An optional sequence of torch modules containing transforms to be applied on the
        input **before** it is fed into the network.
    pretrained
        If True, will use VGG16 pretrained weights.
    """

    def __init__(
        self,
        loss_type: type[torch.nn.Module] = SoftJaccardAndBCEWithLogitsLoss,
        loss_arguments: dict[str, typing.Any] | None = None,
        optimizer_type: type[torch.optim.Optimizer] = torch.optim.Adam,
        optimizer_arguments: dict[str, typing.Any] | None = None,
        scheduler_type: type[torch.optim.lr_scheduler.LRScheduler] | None = None,
        scheduler_arguments: dict[str, typing.Any] | None = None,
        model_transforms: TransformSequence | None = None,
        augmentation_transforms: TransformSequence | None = None,
        pretrained: bool = False,
    ):
        super().__init__(
            name="driu-od",
            loss_type=loss_type,
            loss_arguments=loss_arguments,
            optimizer_type=optimizer_type,
            optimizer_arguments=optimizer_arguments,
            scheduler_type=scheduler_type,
            scheduler_arguments=scheduler_arguments,
            model_transforms=model_transforms,
            augmentation_transforms=augmentation_transforms,
            num_classes=1,  # fixed at current implementation
        )

        if pretrained:
            from ..normalizer import make_imagenet_normalizer

            Model.normalizer.fset(self, make_imagenet_normalizer())  # type: ignore[attr-defined]

        self.backbone = vgg16_for_segmentation(
            pretrained=pretrained,
            return_features=[8, 14, 22, 29],
        )

        self.head = DRIUODHead([128, 256, 512, 512])

    def forward(self, x):
        x = self.normalizer(x)
        x = self.backbone(x)
        return self.head(x)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import typing
from functools import singledispatchmethod

import numpy
import torch
import torch.nn
import torchvision.transforms.v2
import torchvision.transforms.v2.functional
import torchvision.tv_tensors


def crop_image_to_mask(img: torch.Tensor, mask: torch.Tensor) -> torch.Tensor:
    """Square crop image to the boundaries of a boolean mask.

    Parameters
    ----------
    img
        The image to crop, of shape channels x height x width.
    mask
        The boolean mask to use for cropping.

    Returns
    -------
        The cropped image.
    """

    if img.shape[-2:] != mask.shape[-2:]:
        raise ValueError(
            f"Image and mask must have the same size: {img.shape[-2:]} != {mask.shape[-2:]}"
        )

    h, w = img.shape[-2:]

    flat_mask = mask.flatten()
    top = flat_mask.nonzero()[0] // w
    bottom = h - (torch.flip(flat_mask, dims=(0,)).nonzero()[0] // w)

    flat_transposed_mask = torch.transpose(mask, 1, 2).flatten()
    left = flat_transposed_mask.nonzero()[0] // h
    right = w - (torch.flip(flat_transposed_mask, dims=(0,)).nonzero()[0] // h)

    return img[:, top:bottom, left:right]


def square_center_pad(img: torch.Tensor, size: typing.Any) -> torch.Tensor:
    """Return a squared version of the image, centered on a canvas padded with
    zeros.

    Parameters
    ----------
    img
        The tensor to be transformed.  Expected to be in the form: ``[...,
        [1,3], H, W]`` (i.e. arbitrary number of leading dimensions).
    size
        Height and width of the image.

    Returns
    -------
        Transformed tensor, guaranteed to be square (ie. equal height and
        width).
    """

    height, width = size
    maxdim = numpy.max([height, width])

    # padding
    left = (maxdim - width) // 2
    top = (maxdim - height) // 2
    right = maxdim - width - left
    bottom = maxdim - height - top

    return torchvision.transforms.v2.functional.pad(
        img,
        [left, top, right, bottom],
        0,
        "constant",
    )


class SquareCenterPad(torchvision.transforms.v2.Transform):
    """Transform to a squared version of the image, centered on a canvas padded
    with zeros.
    """

    def __init__(self):
        super().__init__()

    @singledispatchmethod
    def _transform(self, inpt: typing.Any, params: dict[str, typing.Any]) -> typing.Any:
        return inpt

    @_transform.register(torch.Tensor)
    @_transform.register(torchvision.tv_tensors.Image)
    @_transform.register(torchvision.tv_tensors.Mask)
    def _(
        self,
        inpt: torch.Tensor | torchvision.tv_tensors.Image | torchvision.tv_tensors.Mask,
        params: dict[str, typing.Any],
    ) -> torch.Tensor | torchvision.tv_tensors.Image | torchvision.tv_tensors.Mask:
        return square_center_pad(inpt, inpt.shape[-2:])

    @_transform.register(torchvision.tv_tensors.BoundingBoxes)
    def _(
        self, inpt: torchvision.tv_tensors.BoundingBoxes, params: dict[str, typing.Any]
    ) -> torchvision.tv_tensors.BoundingBoxes:
        return square_center_pad(inpt, inpt.canvas_size)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Color/grayscale jitter 20% transformations."""

import torchvision.transforms

augmentations = [
    torchvision.transforms.ColorJitter(
        brightness=0.2,
        contrast=0.2,
        saturation=0.2,
    )
]



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Horizontal flip with 50% probability."""

import torchvision.transforms

augmentations = [torchvision.transforms.RandomHorizontalFlip(p=0.5)]



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tests for the aggregated Montgomery-Shenzhen-Indian-TBX11k dataset."""

import importlib

import pytest
from click.testing import CliRunner


@pytest.mark.parametrize(
    "name,tbx11k_name",
    [
        ("default", "v1_healthy_vs_atb"),
        ("fold_0", "v1_fold_0"),
        ("fold_1", "v1_fold_1"),
        ("fold_2", "v1_fold_2"),
        ("fold_3", "v1_fold_3"),
        ("fold_4", "v1_fold_4"),
        ("fold_5", "v1_fold_5"),
        ("fold_6", "v1_fold_6"),
        ("fold_7", "v1_fold_7"),
        ("fold_8", "v1_fold_8"),
        ("fold_9", "v1_fold_9"),
        ("default", "v2_others_vs_atb"),
        ("fold_0", "v2_fold_0"),
        ("fold_1", "v2_fold_1"),
        ("fold_2", "v2_fold_2"),
        ("fold_3", "v2_fold_3"),
        ("fold_4", "v2_fold_4"),
        ("fold_5", "v2_fold_5"),
        ("fold_6", "v2_fold_6"),
        ("fold_7", "v2_fold_7"),
        ("fold_8", "v2_fold_8"),
        ("fold_9", "v2_fold_9"),
    ],
)
def test_split_consistency(name: str, tbx11k_name: str):
    montgomery = importlib.import_module(
        f".{name}", "mednet.config.classify.data.montgomery"
    ).datamodule

    shenzhen = importlib.import_module(
        f".{name}", "mednet.config.classify.data.shenzhen"
    ).datamodule

    indian = importlib.import_module(
        f".{name}", "mednet.config.classify.data.indian"
    ).datamodule

    tbx11k = importlib.import_module(
        f".{tbx11k_name}", "mednet.config.classify.data.tbx11k"
    ).datamodule

    combined = importlib.import_module(
        f".{tbx11k_name}",
        "mednet.config.classify.data.montgomery_shenzhen_indian_tbx11k",
    ).datamodule

    montgomery_loader = importlib.import_module(
        ".montgomery", "mednet.data.classify"
    ).RawDataLoader

    shenzhen_loader = importlib.import_module(
        ".shenzhen", "mednet.data.classify"
    ).RawDataLoader

    indian_loader = importlib.import_module(
        ".indian", "mednet.data.classify"
    ).RawDataLoader

    tbx11k_loader = importlib.import_module(
        ".tbx11k", "mednet.data.classify"
    ).RawDataLoader

    for split in ("train", "validation", "test"):
        assert montgomery.splits[split][0][0] == combined.splits[split][0][0]
        assert isinstance(montgomery.splits[split][0][1], montgomery_loader)
        assert isinstance(combined.splits[split][0][1], montgomery_loader)

        assert shenzhen.splits[split][0][0] == combined.splits[split][1][0]
        assert isinstance(shenzhen.splits[split][0][1], shenzhen_loader)
        assert isinstance(combined.splits[split][1][1], shenzhen_loader)

        assert indian.splits[split][0][0] == combined.splits[split][2][0]
        assert isinstance(indian.splits[split][0][1], indian_loader)
        assert isinstance(combined.splits[split][2][1], indian_loader)

        assert tbx11k.splits[split][0][0] == combined.splits[split][3][0]
        assert isinstance(tbx11k.splits[split][0][1], tbx11k_loader)
        assert isinstance(combined.splits[split][3][1], tbx11k_loader)


@pytest.mark.parametrize(
    "dataset",
    [
        "train",
    ],
)
@pytest.mark.parametrize(
    "tbx11k_name",
    [
        ("v1_healthy_vs_atb"),
    ],
)
def test_batch_uniformity(tbx11k_name: str, dataset: str):
    combined = importlib.import_module(
        f".{tbx11k_name}",
        "mednet.config.classify.data.montgomery_shenzhen_indian_tbx11k",
    ).datamodule

    combined.model_transforms = []  # should be done before setup()
    combined.setup("predict")  # sets up all datasets

    loader = combined.predict_dataloader()[dataset]

    limit = 5  # limit load checking
    for batch in loader:
        if limit == 0:
            break
        assert len(batch) == 3  # image, target, name


@pytest.mark.slow
@pytest.mark.skip_if_rc_var_not_set("datadir.montgomery")
@pytest.mark.skip_if_rc_var_not_set("datadir.shenzhen")
@pytest.mark.skip_if_rc_var_not_set("datadir.indian")
@pytest.mark.skip_if_rc_var_not_set("datadir.tbx11k")
def test_database_check():
    from mednet.scripts.database import check

    runner = CliRunner()
    result = runner.invoke(check, ["montgomery-shenzhen-indian-tbx11k-v1"])
    assert (
        result.exit_code == 0
    ), f"Exit code {result.exit_code} != 0 -- Output:\n{result.output}"



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Support code to adapt existing `MobileNetV2 pytorch model <mobilenetv2-pytorch_>`_ functionality to U-Net style network for segmentation."""

import torchvision.models
import torchvision.models.mobilenetv2
from torch.hub import load_state_dict_from_url


class MobileNetV24Segmentation(torchvision.models.mobilenetv2.MobileNetV2):
    """Adaptation of `MobileNetV2 pytorch model <mobilenetv2-pytorch_>`_ to U-Net style network for segmentation.

    This version of MobileNetV2 is slightly modified so it can be used through
    torchvision's API.  It outputs intermediate features which are normally not
    output by the base MobileNetV2 implementation, but are required for
    segmentation operations.

    Parameters
    ----------
    *args
        Arguments to be passed to the parent MobileNetV2 model.
    **kwargs
        Keyword arguments to be passed to the parent MobileNetV2 model.

        * ``return_features`` (:py:class:`list`): An optional list of integers indicating
          the feature layers to be returned from the original module.
    """

    def __init__(self, *args, **kwargs):
        self._return_features = kwargs.pop("return_features")
        super().__init__(*args, **kwargs)

    def forward(self, x):
        outputs = []
        # hw of input, needed for DRIU and HED
        outputs.append(x.shape[2:4])
        outputs.append(x)
        for index, m in enumerate(self.features):
            x = m(x)
            # extract layers
            if index in self._return_features:
                outputs.append(x)
        return outputs


def mobilenet_v2_for_segmentation(pretrained=False, progress=True, **kwargs):
    """Create MobileNetV2 model for segmentation task.

    Parameters
    ----------
    pretrained
        If True, uses MobileNetV2 pretrained weights.
    progress
        If True, shows a progress bar when downloading the pretrained weights.
    **kwargs
        Keyword arguments to be passed to the parent MobileNetV2 model.

        * ``return_features`` (:py:class:`list`): An optional list of integers indicating
          the feature layers to be returned from the original module.

    Returns
    -------
        Instance of the MobileNetV2 model for segmentation.
    """

    model = MobileNetV24Segmentation(**kwargs)

    if pretrained:
        state_dict = load_state_dict_from_url(
            torchvision.models.mobilenetv2.MobileNet_V2_Weights.DEFAULT.url,
            progress=progress,
        )
        model.load_state_dict(state_dict)

    # erase MobileNetV2 head (for classification), not used for segmentation
    delattr(model, "classifier")

    return_features = kwargs.get("return_features")
    if return_features is not None:
        model.features = model.features[: (max(return_features) + 1)]

    return model


mobilenet_v2_for_segmentation.__doc__ = torchvision.models.mobilenetv2.__doc__
